import numpy as np
from pp.routing.route_ports_to_side import route_ports_to_side
from pp.routing.connect import connect_strip_way_points
from pp.routing.manhattan import generate_manhattan_waypoints
from pp.routing.manhattan import remove_flat_angles
from pp.geo_utils import remove_identicals


def _groups(ports, cut, axis="X"):
    if axis == "Y":
        group1 = [p for p in ports if p.x <= cut]
        group2 = [p for p in ports if p.x > cut]
    else:
        group1 = [p for p in ports if p.y <= cut]
        group2 = [p for p in ports if p.y > cut]
    return group1, group2


def u_bundle_direct(
    start_ports,
    end_ports,
    route_filter=connect_strip_way_points,
    separation=5.0,
    start_straight=0.01,
    end_straight=0.01,
    start_straight_offset=0.0,
    end_straight_offset=0.0,
    **routing_params
):
    r"""
    Args:
        start_ports: list of start ports
        end_ports: list of end ports
        route_filter: filter to apply to the manhattan waypoints
            e.g `connect_strip_way_points` for deep etch strip waveguide

    Returns:
        `[route_filter(r) for r in routes]` where routes is a list of lists of coordinates
        e.g with default `connect_strip_way_points`, returns a list of elements which can be added to a component


    Used for routing multiple ports back to a bundled input in a component

    ::

        X: start ports
        D: End ports

        On this example bellow, the axis is along X

                           X------\
                                  |
                           X----\ |
                          ...   | |
                           X--\ | |
                              | | |
          D-------------------/ | |
          ...                   | |
          D---------------------/ |
          D-----------------------
          D-----------------------\
          D---------------------\ |
          ...                   | |
          D-------------------\ | |
                              | | |
                           X--/ | |
                          ...   | |
                           X----/ |
                                  |
                           X------/
    """

    routes = u_bundle_direct_routes(
        start_ports,
        end_ports,
        separation=separation,
        start_straight=start_straight,
        end_straight=end_straight,
        start_straight_offset=start_straight_offset,
        end_straight_offset=end_straight_offset,
        routing_func=generate_manhattan_waypoints,
        **routing_params
    )

    return [route_filter(route, **routing_params) for route in routes]


def u_bundle_direct_routes(
    start_ports,
    end_ports,
    routing_func=generate_manhattan_waypoints,
    separation=5.0,
    start_straight=0.01,
    end_straight=0.01,
    end_straight_offset=0.0,
    start_straight_offset=0.0,
    **routing_func_params
):

    nb_ports = len(start_ports)
    for p in start_ports:
        p.angle = p.angle % 360

    for p in end_ports:
        p.angle = p.angle % 360

    if len(end_ports) != nb_ports:
        raise ValueError(
            "Number of start ports should match number of end ports.\
        Got {} {}".format(
                len(start_ports), len(end_ports)
            )
        )
    if len(set([p.angle for p in start_ports + end_ports])) > 1:
        raise ValueError(
            "All ports should have the same angle\
        , got \n{}\n{}".format(
                start_ports, end_ports
            )
        )

    xs_end = [p.x for p in end_ports]
    ys_end = [p.y for p in end_ports]

    x_cut = 0.5 * (min(xs_end) + max(xs_end))
    y_cut = 0.5 * (min(ys_end) + max(ys_end))

    # Find axis
    angle_start = start_ports[0].angle
    if angle_start in [0, 180]:
        axis = "X"
        cut = y_cut
    else:
        axis = "Y"
        cut = x_cut

    # Get groups (below, above) or (left, right)
    group1, group2 = _groups(start_ports, cut, axis=axis)

    # Sort ports to make them easy to connect
    if axis == "X":
        group1.sort(key=lambda p: -p.y)
        group2.sort(key=lambda p: p.y)

        end_ports.sort(key=lambda p: p.y)
        end_group1 = end_ports[: len(group1)]
        end_group2 = end_ports[len(group1) :]
        end_group2.sort(key=lambda p: -p.y)

        xs_start = [p.x for p in start_ports]

        if angle_start == 0:
            dx = xs_start[0] - xs_end[0]
        elif angle_start == 180:
            dx = xs_end[0] - xs_start[0]
        end_straight = max(end_straight, dx)

    if axis == "Y":
        group1.sort(key=lambda p: -p.x)
        group2.sort(key=lambda p: p.x)

        end_ports.sort(key=lambda p: p.x)
        end_group1 = end_ports[: len(group1)]
        end_group2 = end_ports[len(group1) :]
        end_group2.sort(key=lambda p: -p.x)

        ys_start = [p.y for p in start_ports]

        if angle_start == 90:
            dy = ys_start[0] - ys_end[0]
        elif angle_start == 270:
            dy = ys_end[0] - ys_start[0]
        end_straight = max(end_straight, dy)

    # add offsets
    start_straight += start_straight_offset
    end_straight += end_straight_offset

    connections = []
    straight_len_end = end_straight
    straight_len_start = start_straight
    for p_start, p_end in zip(group1, end_group1):
        _c = routing_func(
            p_start,
            p_end,
            start_straight=straight_len_start,
            end_straight=straight_len_end,
            **routing_func_params
        )
        connections += [_c]
        straight_len_end += separation
        straight_len_start += separation

    straight_len_end = end_straight
    straight_len_start = start_straight
    for p_start, p_end in zip(group2, end_group2):
        _c = routing_func(
            p_start,
            p_end,
            start_straight=straight_len_start,
            end_straight=straight_len_end,
            **routing_func_params
        )
        connections += [_c]
        straight_len_end += separation
        straight_len_start += separation

    return connections


def u_bundle_indirect(
    start_ports,
    end_ports,
    route_filter=connect_strip_way_points,
    separation=5.0,
    extension_length=0.0,
    start_straight=0.01,
    end_straight=0.01,
    end_straight_offset=0.0,
    start_straight_offset=0.0,
    **routing_params
):
    """
    Args:
        start_ports: list of start ports
        end_ports: list of end ports
        route_filter: filter to apply to the manhattan waypoints
            e.g `connect_strip_way_points` for deep etch strip waveguide
    Returns:
        `[route_filter(r) for r in routes]` where routes is a list of lists of coordinates
        e.g with default `connect_strip_way_points`, returns a list of elements which can be added to a component


    Used for routing multiple ports back to a bundled input in a component


    ::

        X: start ports
        D: End ports


                                    X------
                                    X----  |
                                   ...   | |
                                    X--\ | |
                                       | | |
            /--------------------------/ | |
            |                            | |
            | /--------------------------/ |
            | |                            |
            | | /--------------------------/
            | | |
            | | \--D
            | \----D
            |     ...
            \------D
            /------D
            |     ...
            | /----D
            | | /--D
            | | |
            | | \--------------------------
            | |                            |
            | \--------------------------\ |
            |                            | |
            \--------------------------\ | |
                                       | | |
                                    X--/ | |
                                   ...   | |
                                    X----/ |
                                    X------/
        '''

    """

    routes = u_bundle_indirect_routes(
        start_ports,
        end_ports,
        separation=separation,
        start_straight=start_straight,
        end_straight=end_straight,
        start_straight_offset=start_straight_offset,
        end_straight_offset=end_straight_offset,
        routing_func=generate_manhattan_waypoints,
        extension_length=extension_length,
        **routing_params
    )

    return [route_filter(route, **routing_params) for route in routes]


def u_bundle_indirect_routes(
    start_ports,
    end_ports,
    routing_func=generate_manhattan_waypoints,
    separation=5.0,
    extension_length=0.0,
    start_straight=0.01,
    end_straight=0.01,
    end_straight_offset=0.0,
    start_straight_offset=0.0,
    **routing_func_params
):

    nb_ports = len(start_ports)

    for p in start_ports:
        p.angle = p.angle % 360

    for p in end_ports:
        p.angle = p.angle % 360

    if len(end_ports) != nb_ports:
        raise ValueError(
            "Number of start ports should match number of end ports.\
        Got {} {}".format(
                len(start_ports), len(end_ports)
            )
        )

    if len(set([p.angle for p in start_ports])) > 1:
        raise ValueError(
            "All start port angles should be the same.\
        Got {}".format(
                start_ports
            )
        )

    if len(set([p.angle for p in end_ports])) > 1:
        raise ValueError(
            "All end port angles should be the same.\
        Got {}".format(
                end_ports
            )
        )

    xs_end = [p.x for p in end_ports]
    ys_end = [p.y for p in end_ports]

    """
    # Compute the bundle axis
    """

    if start_ports[0].angle in [0, 180]:
        axis = "X"
    else:
        axis = "Y"

    """
     Split start ports in two groups:
        - the ones on the south/west of end ports (depending on bundle axis)
        - the ones on the north/east of end ports (depending on bundle axis)
    """

    if axis == "X":
        y_cut = 0.5 * (min(ys_end) + max(ys_end))
        group1 = [p for p in start_ports if p.y <= y_cut]
        group2 = [p for p in start_ports if p.y > y_cut]

        if start_ports[0].angle == 0 and end_ports[0].angle == 180:
            """
                    X->
               <-D
                    X->
            """
            # To go back to a U bundle
            group1_route_directives = ["north", "west"]
            group2_route_directives = ["south", "west"]

        elif start_ports[0].angle == 180 and end_ports[0].angle == 0:
            """
               <-X
                    D->
               <-X
            """
            # To go back to a U bundle
            group1_route_directives = ["north", "east"]
            group2_route_directives = ["south", "east"]

        else:
            print("u_undirect_bundle not designed to work in this case")

    if axis == "Y":
        x_cut = 0.5 * (min(xs_end) + max(xs_end))
        group1 = [p for p in start_ports if p.x <= x_cut]
        group2 = [p for p in start_ports if p.x > x_cut]

        if start_ports[0].angle == 90 and end_ports[0].angle == 270:
            """

              ^     ^
              |     |
              X     X
                 D
                 |

            """
            # To go back to a U bundle
            group1_route_directives = ["east", "south"]
            group2_route_directives = ["west", "south"]

        elif start_ports[0].angle == 270 and end_ports[0].angle == 90:
            """
                 ^
                 |
                 D
              X     X
              |     |

            """
            # To go back to a U bundle
            group1_route_directives = ["east", "north"]
            group2_route_directives = ["west", "north"]

        else:
            print("u_undirect_bundle not designed to work in this case")

    """
    Do the routing directives to get back to a u_bundle direct case
    """

    routing_param = {
        "routing_func": routing_func,
        "separation": separation,
        **routing_func_params,
    }

    # Multiple sections of different routes are generated in different places.
    # At the output, we want a list of routes. (not a list of portions of route)
    # dict_connections keeps track of these sections

    dict_connections = {i: [] for i in range(nb_ports)}

    def add_connections(conns):
        """
        Ensure that each section in a batch of connection
        is added to the correct route. Also we don't know in which order the
        routes are given (from beginning to end or vice versa)
        """

        end_prev_conns = [(k, v[-1][-1]) for k, v in dict_connections.items()]
        for c in conns:
            p = c[0]
            for i, q in end_prev_conns:
                if np.abs(p - q).sum() < 1e-9:
                    dict_connections[i] += [c]
                    break

    # First part
    print(group1_route_directives)
    conn1, tmp_ports1 = route_ports_to_side(
        group1,
        group1_route_directives[0],
        extension_length=extension_length,
        **routing_param
    )

    conn2, tmp_ports2 = route_ports_to_side(
        group2,
        group2_route_directives[0],
        extension_length=extension_length,
        **routing_param
    )
    conn = conn1 + conn2
    dict_connections = {i: [c] for i, c in enumerate(conn)}

    # Second part
    conn1, tmp_ports1 = route_ports_to_side(
        tmp_ports1, group1_route_directives[1], **routing_param
    )

    conn2, tmp_ports2 = route_ports_to_side(
        tmp_ports2, group2_route_directives[1], **routing_param
    )

    add_connections(conn1 + conn2)
    # conn2)

    bundle_params = {
        **routing_param,
        "start_straight": start_straight,
        "end_straight": end_straight,
        "start_straight_offset": start_straight_offset,
        "end_straight_offset": end_straight_offset,
    }

    end_ports.sort(key=lambda p: p.y)
    conn1 = u_bundle_direct_routes(
        tmp_ports1, end_ports[: len(tmp_ports1)], **bundle_params
    )
    conn2 = u_bundle_direct_routes(
        tmp_ports2, end_ports[len(tmp_ports1) :], **bundle_params
    )

    add_connections(conn1 + conn2)

    def _merge_connections(list_of_points):

        a = [list_of_points[0]]
        a = a + [point[1:] for point in list_of_points[1:]]
        b = np.vstack(a)
        b = remove_identicals(b)
        b = remove_flat_angles(b)
        return b

    connections = [_merge_connections(c) for c in dict_connections.values()]
    return connections



""" based on https://github.com/niladri18/Phidl/blob/master/src/ppe.py
"""

import math
import pp
from omegaconf.listconfig import ListConfig
from pp.component import Component
from typing import List, Union


@pp.autoname
def line(
    x0: Union[float, int],
    y0: Union[float, int],
    width: Union[float, int],
    height: int,
    layer: ListConfig,
) -> Component:
    L = pp.Component()
    L.add_polygon(
        [(x0, y0), (x0 + width, y0), (x0 + width, y0 + height), (x0, y0 + height)],
        layer=layer,
    )
    return L


@pp.autoname
def linespace(
    x0: Union[float, int],
    y0: Union[float, int],
    width: Union[float, int],
    height: int,
    pitch: int,
    ymax: Union[float, int],
    layer: ListConfig,
) -> Component:
    """ # Creates a line space pattern in y-direction
    Args:
        x0: x coordinate of the lower left line
        y0: y coordinate of the lower left line
        width: width of each line
        height: height of each line
        pitch: pitch of each line
        pitch > height
    """
    if abs(pitch) < abs(height):
        print("pitch must be greater then height")
        return
    LS = pp.Component()
    if pitch > 0:
        while y0 + height <= ymax:
            Li = line(x0=x0, y0=y0, width=width, height=height, layer=layer)
            LS.add_ref(Li)
            y0 += pitch
    elif pitch < 0:
        while y0 + height >= -ymax:
            Li = line(x0=x0, y0=y0, width=width, heigh=height, layer=layer)
            LS.add_ref(Li)
            y0 += pitch
    return LS


def y0linespace(y0: float, height: int, pitch: int, ymax: float) -> float:
    if pitch > 0:
        while y0 + height <= ymax:
            y0 += pitch
    elif pitch < 0:
        while y0 + height >= -ymax:
            y0 += pitch
    return y0


@pp.autoname
def cross(x0: float, y0: float, width: int, lw: int, layer: ListConfig) -> Component:
    """ cross

    Args:
        x0,y0 : center
        width: width of the bounding box
        lw: linewidth
    """
    cross = pp.Component()
    cross.add_polygon(
        [
            (x0 - width / 2, y0 - lw / 2),
            (x0 - width / 2, y0 + lw / 2),
            (x0 + width / 2, y0 + lw / 2),
            (x0 + width / 2, y0 - lw / 2),
        ],
        layer=layer,
    )
    cross.add_polygon(
        [
            (x0 - lw / 2, y0 - width / 2),
            (x0 - lw / 2, y0 + width / 2),
            (x0 + lw / 2, y0 + width / 2),
            (x0 + lw / 2, y0 - width / 2),
        ],
        layer=layer,
    )
    return cross


@pp.autoname
def ppe(
    layer: ListConfig = pp.LAYER.WG,
    layers_cladding: List[ListConfig] = [pp.LAYER.WGCLAD],
    cladding_offset: int = 3,
) -> Component:
    """
    pattern placement error
    """
    D = pp.Component()

    # Define global variables
    xmax = 500
    ymax = 500
    xmin = 0
    ymin = 0
    xm = (xmax - xmin) / 2.0
    ym = (ymax - ymin) / 2.0
    o = cladding_offset

    # Cover the entire macro
    for layer_cladding in layers_cladding:
        D.add_polygon(
            [(0, -o), (xmax + o, -o), (xmax + o, ymax + o), (0, ymax + o)],
            layer=layer_cladding,
        )

    # Place the pattern rec
    Cross = cross(x0=xm, y0=ym, width=100, lw=10, layer=layer)
    Cross.rotate(45, center=[xm, ym])
    D.add_ref(Cross)

    # calculate offset due to the cross
    xoff = math.sqrt(100 * 50)
    yoff = math.sqrt(100 * 50)

    # Top left 1
    x0 = 10
    y0 = ym
    pitch = 20
    LS1 = linespace(
        x0=x0,
        y0=y0,
        width=240 - xoff,
        height=10,
        pitch=pitch,
        ymax=ym + yoff,
        layer=layer,
    )
    y0 = y0linespace(y0=y0, height=10, pitch=pitch, ymax=ym + yoff)
    D.add_ref(LS1)

    # Top left 2
    x0 = 10
    y0 = y0
    LS1 = linespace(x0=x0, y0=y0, width=240, height=10, pitch=20, ymax=500, layer=layer)
    D.add_ref(LS1)

    # Top right 1
    x0 = xm + xoff
    y0 = ym
    pitch = 30
    LS2 = linespace(
        x0=x0,
        y0=y0,
        width=240 - xoff + 10,
        height=10,
        pitch=pitch,
        ymax=ym + yoff,
        layer=layer,
    )
    y0 = y0linespace(y0=y0, height=10, pitch=pitch, ymax=ym + yoff)
    D.add_ref(LS2)

    # Top right 2
    x0 = xm + 10
    LS2 = linespace(x0=x0, y0=y0, width=240, height=10, pitch=30, ymax=500, layer=layer)
    D.add_ref(LS2)

    # Lower left 1
    x0 = 10
    y0 = 0
    pitch = 30
    LS3 = linespace(
        x0=x0, y0=y0, width=240, height=10, pitch=30, ymax=xm - yoff, layer=layer
    )
    D.add_ref(LS3)

    # Lower left 2
    x0 = 10
    y0 += pitch
    LS3 = linespace(
        x0=x0, y0=y0, width=240 - xoff, height=10, pitch=30, ymax=240, layer=layer
    )
    D.add_ref(LS3)

    # Lower right 1
    x0 = xm + 10
    y0 = 0
    pitch = 20
    LS4 = linespace(
        x0=x0, y0=y0, width=240, height=10, pitch=20, ymax=xm - yoff, layer=layer
    )
    D.add_ref(LS4)

    # Lower right 2
    x0 = xm + xoff
    y0 += pitch
    LS4 = linespace(
        x0=x0, y0=y0, width=240 - xoff + 10, height=10, pitch=20, ymax=240, layer=layer
    )
    D.add_ref(LS4)

    # Add NOOPC cover on the pattern rec
    xt = xoff - 10
    yt = yoff - 10

    for layer_cladding in layers_cladding:
        D.add_polygon(
            [
                (xm - xt, ym - yt),
                (xm - xt, ym + yt),
                (xm + xt, ym + yt),
                (xm + xt, ym - yt),
            ],
            layer=layer_cladding,
        )
    return D


if __name__ == "__main__":
    c = ppe()
    pp.show(c)



from pp.component import Component
from pp.container import container
from pp.components.electrical.pad import pad_array
from pp.routing.connect_electrical import connect_electrical_shortest_path


@container
def add_electrical_pads_top(component, **kwargs):
    """connects component electrical ports with pad array at the top

    Args:
        component:
        pad: pad element
        spacing: pad array (x, y) spacing
        width: pad width
        height: pad height
        layer: pad layer
    """
    c = Component(f"{component}_e")
    ports = component.get_ports_list(port_type="dc")
    c << component
    pads = c << pad_array(n=len(ports), port_list=["S"], **kwargs)
    pads.x = component.x
    pads.y = component.ymax + 100
    ports_pads = list(pads.ports.values())
    for p1, p2 in zip(ports_pads, ports):
        c.add(connect_electrical_shortest_path(p1, p2))

    c.ports = component.ports
    for port in ports:
        c.ports.pop(port.name)
    return c


if __name__ == "__main__":
    import pp

    c = pp.c.mzi2x2(with_elec_connections=True)
    cc = add_electrical_pads_top(c)



from pp.name import autoname
from pp.components import rectangle

import pp
from pp.component import Component


@autoname
def litho_star(
    num_lines: int = 20, line_width: int = 2, diameter: int = 200, layer: int = 0
) -> Component:
    """ Creates a circular-star shape from lines, used as a lithographic
    resolution test pattern

    .. plot::
      :include-source:

      import pp

      c = pp.c.litho_star()
      pp.plotgds(c)
    """
    D = pp.Component()

    degree = 180 / num_lines
    R1 = rectangle(size=(line_width, diameter), layer=layer)
    for i in range(num_lines):
        r1 = D.add_ref(R1).rotate(degree * i)
        r1.center = (0, 0)

    return D


if __name__ == "__main__":
    c = litho_star()
    pp.show(c)



import sys
import numpy as np
import gdspy as gp
from pp.geo_utils import area


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def compute_area(c, target_layer):
    """
    Compute area of the component on a given layer
    """
    _print("Computing area ", c.name)
    c.flatten()
    # return c.area(by_spec=True)[layer]
    polys_by_spec = c.get_polygons(by_spec=True)
    _area = 0
    for (layer, polys) in polys_by_spec.items():
        _print(layer)
        if layer == target_layer:
            joined_polys = gp.boolean(polys, None, operation="or")
            _print(joined_polys)
            try:
                _area += sum([abs(area(p)) for p in joined_polys.polygons])
            except:
                print(
                    "Warning, {} joinedpoly {} could not be added".format(
                        c.name, joined_polys
                    )
                )
    return _area


def bucket_cells_by_rank(cells):
    cells = list(cells)
    rank = 0
    rank_to_cells = {}
    all_classified_cells = set()
    prev_len_cells = -1
    while cells:
        classified_cells = set()
        to_rm = []
        for i, c in enumerate(cells):
            _cells = c.get_dependencies(recursive=False)
            unclassified_subcells = _cells - all_classified_cells
            if len(unclassified_subcells) == 0:
                classified_cells.update([c])
                to_rm += [i]

        if prev_len_cells == len(cells):
            print(cells)
            raise ValueError("Error: some cells cannot be linked")
        prev_len_cells = len(cells)
        while to_rm:
            cells.pop(to_rm.pop())

        rank_to_cells[rank] = classified_cells
        all_classified_cells.update(classified_cells)
        rank += 1
    return rank_to_cells


def get_polygons_on_layer(c, layer):
    polygons = []
    for polyset in c.polygons:
        for ii in range(len(polyset.polygons)):
            key = (polyset.layers[ii], polyset.datatypes[ii])
            if key == layer:
                polygons.append(polyset.polygons[ii])
    return polygons


def boolops_hierarchical(
    c, layer1, layer2, layer_result, operation="or", func_check_to_flatten=None
):

    all_cells = c.get_dependencies(recursive=True)
    all_cells.update([c])
    cells_by_rank = bucket_cells_by_rank(all_cells)
    _print("Found the hierarchy...")

    if func_check_to_flatten is None:

        def _has_polygons(cell):
            n = 0
            for _layer in [layer1, layer2]:
                polys = get_polygons_on_layer(cell, _layer)
                n += len(polys)
            return n

        func_check_to_flatten = _has_polygons

    for rank, cells in cells_by_rank.items():
        for cell in cells:
            to_flatten = func_check_to_flatten(cell)
            if to_flatten:
                _print("BOOL HIERARCHY", cell.name, "...", end="")
                polys_by_spec = cell.get_polygons(by_spec=True)
                polys1 = polys_by_spec[layer1] if layer1 in polys_by_spec else []
                polys2 = polys_by_spec[layer2] if layer2 in polys_by_spec else []
                cell.remove_layers([layer_result])

                res_polys = gp.boolean(polys1, polys2, operation=operation)
                if res_polys is not None:
                    cell.add_polygon(res_polys, layer=layer_result)

                _print("{} - done".format(res_polys), cell.name)

    return cell


def compute_area_hierarchical(
    c, layer, func_check_to_flatten=None, keep_zero_area_cells=False
):
    """
    Compute area of the component on a given layer
    Faster than `compute_area` but need to be careful if the cells overlap
    Can pass a list of cells to flatten
    """

    all_cells = c.get_dependencies(recursive=True)
    all_cells.update([c])
    cells_by_rank = bucket_cells_by_rank(all_cells)
    _print("Found the hierarchy...")

    cell_to_area = {}
    cell_to_rank = {}

    if func_check_to_flatten is None:

        def _has_polygons(cell):
            polys = get_polygons_on_layer(cell, layer)
            return len(polys)

        func_check_to_flatten = _has_polygons

    for rank, cells in cells_by_rank.items():
        for cell in cells:
            to_flatten = func_check_to_flatten(cell)
            if to_flatten:
                _print("CAH - TO FLATTEN", to_flatten)
                _area = compute_area(cell, layer)
            else:
                # _cell_area_by_spec = cell.area(by_spec=True)
                _area = 0
                # _cell_area_by_spec[layer] if layer in _cell_area_by_spec else 0
                # _area = 0

                # _print("CAH - ",cell.name)

                for ref in cell.references:
                    _area += cell_to_area[ref.ref_cell.name]
                _print(
                    "CAH {} {:.1f} {}".format(cell.name, _area, len(cell.references))
                )

            cell_to_area[cell.name] = _area
            cell_to_rank[cell.name] = rank
    to_rm = []
    if not keep_zero_area_cells:
        for k, v in cell_to_area.items():
            if v == 0:
                to_rm += [k]

    while to_rm:
        cell_to_area.pop(to_rm.pop())

    # add rank
    cell_to_data = {}
    for k, v in cell_to_area.items():
        cell_to_data[k] = (v, cell_to_rank[k])

    return cell_to_data


def test_density():
    """
    Return ratio of total area on a layer over the bounding box of the component
    """
    import pp

    layer = (1, 0)
    c = pp.c.rectangle(size=(4, 2), layer=layer)
    pp.write_gds(c)
    area = compute_area(c, layer)
    assert np.isclose(8, area)


if __name__ == "__main__":
    test_density()

    # pp.show(c)

    # c = pp.c.mzi2x2()
    # print(bucket_cells_by_rank([c] + list(c.get_dependencies(recursive=True))))



from typing import Callable
import pp
from pp.components.bend_circular import bend_circular
from pp.components.coupler90 import coupler90
from pp.components.waveguide import waveguide
from pp.components.coupler_straight import coupler_straight
from pp.netlist_to_gds import netlist_to_component
from pp.name import autoname
from pp.drc import assert_on_2nm_grid
from pp.component import Component


@autoname
def ring_single_bus_deprecated(**kwargs):
    """ Ring single bus

    use single_bus instead as this one can have snaping issues that create gaps between waveguides of the ring

    Args:
        bend_radius=5
        length_x=1
        length_y=1
        gap=0.2
        bend90_factory=bend_circular
        coupler90_factory=coupler90
        straight_factory=waveguide
        cpl_straight_factory=coupler_straight
        wg_width=0.5


    .. code::

         --length_x-
         |         |
         |       length_y
         |         |
       ---===gap===---

    .. plot::
      :include-source:

      import pp

      c = pp.c.ring_single_bus(gap=0.2, length_x=10, length_y=5, bend_radius=5)
      pp.plotgds(c)

    """
    components, connections, ports_map = ring_single_bus_netlist(**kwargs)
    component = netlist_to_component(components, connections, ports_map)
    return component


def ring_single_bus_netlist(
    bend_radius=5,
    length_x=1,
    length_y=1,
    gap=0.2,
    bend90_factory=bend_circular,
    coupler90_factory=coupler90,
    straight_factory=waveguide,
    cpl_straight_factory=coupler_straight,
    wg_width=0.5,
):
    """
    .. code::

         BL--H1--BR
         |       |
         VL      VR
         |       |
        -CL==CS==CR-
    """
    assert_on_2nm_grid(gap)

    bend = bend90_factory(width=wg_width, radius=bend_radius)
    cpl_bend = coupler90_factory(bend_radius=bend_radius, width=wg_width, gap=gap)
    cpl_straight = cpl_straight_factory(length=length_x, gap=gap, width=wg_width)
    h1 = straight_factory(length=length_x, width=wg_width)
    v = straight_factory(length=length_y, width=wg_width)

    components = {
        "CL": (cpl_bend, "mirror_y"),
        "CR": (cpl_bend, "None"),
        "BR": (bend, "mirror_x"),
        "BL": (bend, "R180"),
        "CS": (cpl_straight, "None"),
        "H1": (h1, "None"),
        "VL": (v, "R90"),
        "VR": (v, "R90"),
    }

    connections = [
        ("CL", "W0", "CS", "W0"),
        ("CS", "E0", "CR", "W0"),
        ("CR", "N0", "VR", "W0"),
        ("VR", "E0", "BR", "N0"),
        ("BR", "W0", "H1", "E0"),
        ("H1", "W0", "BL", "W0"),
        ("BL", "N0", "VL", "E0"),
        ("VL", "W0", "CL", "N0"),
    ]

    ports_map = {"W0": ("CL", "E0"), "E0": ("CR", "E0")}

    return components, connections, ports_map


@pp.autoname
def ring_single_bus(
    coupler90_factory: Callable = coupler90,
    cpl_straight_factory: Callable = coupler_straight,
    straight_factory: Callable = waveguide,
    bend90_factory: Callable = bend_circular,
    length_y: float = 2.0,
    length_x: float = 4.0,
    gap: float = 0.2,
    wg_width: float = 0.5,
    bend_radius: int = 5,
) -> Component:
    """ single bus ring
    .. code::

         ctl--wx--ctr
          |       |
         wyl     wgr
          |       |
        -cbl==CS==cbr-
    """
    c = pp.Component()
    assert_on_2nm_grid(gap)

    # define subcells
    coupler90 = pp.call_if_func(
        coupler90_factory, gap=gap, width=wg_width, bend_radius=bend_radius
    )
    waveguide_x = pp.call_if_func(
        straight_factory, length=length_x, width=wg_width, pins=False
    )
    waveguide_y = pp.call_if_func(
        straight_factory, length=length_y, width=wg_width, pins=False
    )
    bend = pp.call_if_func(
        bend90_factory, width=wg_width, radius=bend_radius, pins=False
    )
    coupler_straight = pp.call_if_func(
        cpl_straight_factory, gap=gap, length=length_x, width=wg_width
    )

    # add references to subcells
    cbl = c << coupler90
    cbr = c << coupler90
    cs = c << coupler_straight
    wyl = c << waveguide_y
    wyr = c << waveguide_y
    wx = c << waveguide_x
    btl = c << bend
    btr = c << bend

    # connect references
    wyr.connect(port="E0", destination=cbr.ports["N0"])
    cs.connect(port="E0", destination=cbr.ports["W0"])

    cbl.reflect(p1=(0, coupler90.y), p2=(1, coupler90.y))
    cbl.connect(port="W0", destination=cs.ports["W0"])
    wyl.connect(port="E0", destination=cbl.ports["N0"])

    btl.connect(port="N0", destination=wyl.ports["W0"])
    btr.connect(port="W0", destination=wyr.ports["W0"])
    wx.connect(port="W0", destination=btl.ports["W0"])

    c.add_port("W0", port=cbl.ports["E0"])
    c.add_port("E0", port=cbr.ports["E0"])
    return c


def _compare_rings():

    c = pp.Component()
    c1 = ring_single_bus_deprecated(
        wg_width=0.45, gap=0.15, length_x=0.2, length_y=0.13
    )
    c2 = ring_single_bus(wg_width=0.45, gap=0.15, length_x=0.2, length_y=0.13)

    r1 = c << c1
    r2 = c << c2
    r1.ymin = 0
    r2.ymin = 0
    r1.xmin = 0
    r2.xmin = 0
    pp.show(c)


if __name__ == "__main__":
    # c = ring_single_bus(bend_radius=5.0, length_x=0.2, length_y=0.13, gap=0.15, wg_width=0.45)
    c = ring_single_bus(bend_radius=5.0, gap=0.3, wg_width=0.45)
    print(c.get_settings())
    pp.show(c)
    # c = ring_single_bus(bend_radius=5.0, length_x=2, length_y=4, gap=0.2, wg_width=0.4)
    # _compare_rings()



import pp
from pp.components.grating_coupler.grating_coupler_tree import grating_coupler_tree


def shorted_pads(width=100, n_pads=8, pad_spacing=150, layer=pp.LAYER.WG):
    c = pp.Component(name="shorted_pads")
    pad = pp.c.rectangle_centered(x=width, y=width, layer=layer)
    for i in range(n_pads):
        pad_ref = c.add_ref(pad)
        pad_ref.movex(i * pad_spacing - n_pads / 2 * pad_spacing + pad_spacing / 2)

    short = pp.c.rectangle_centered(x=pad_spacing * (n_pads - 1), y=10, layer=layer)
    c.add_ref(short)
    return c


@pp.autoname
def grating_coupler_tree_with_pads():
    c = pp.Component()

    gratings = c << grating_coupler_tree()
    pads = c << shorted_pads()

    gratings.move(-gratings.size_info.center)
    pads.movey(gratings.ymax + 10 - pads.ymin)
    return c


if __name__ == "__main__":
    # c = shorted_pads()
    c = grating_coupler_tree_with_pads()
    pp.show(c)



from typing import Callable
import pp

from pp.components import bend_circular
from pp.components import wg_heater_connected as waveguide_heater
from pp.components import waveguide
from pp.components import mmi1x2
from pp.components.mzi2x2 import mzi_arm
from pp.netlist_to_gds import netlist_to_component
from pp.routing import route_elec_ports_to_side
from pp.port import select_electrical_ports

from pp.components.extension import line
from pp.component import Component


@pp.autoname
def mzi1x2(
    L0: float = 0.1,
    DL: float = 9.0,
    L2: float = 10.0,
    bend_radius: float = 10.0,
    bend90_factory: Callable = bend_circular,
    straight_heater_factory: Callable = waveguide_heater,
    straight_factory: Callable = waveguide,
    coupler_factory: Callable = mmi1x2,
    with_elec_connections: bool = False,
) -> Component:
    """ Mzi 1x2

    Args:
        L0: vertical length for both and top arms
        DL: bottom arm extra length
        L2: L_top horizontal length
        bend_radius: 10.0
        bend90_factory: bend_circular
        straight_heater_factory: waveguide_heater or waveguide
        straight_factory: waveguide
        coupler_factory: coupler

    .. code::

             __L2__
            |      |
            L0     L0
            |      |
          --|      |--
            |      |
            L0     L0
            |      |
            DL     DL
            |      |
            |__L2__|


             top_arm
        -CP1=       =CP2-
             bot_arm


    .. plot::
      :include-source:

      import pp

      c = pp.c.mzi1x2(L0=0.1, DL=0, L2=10)
      pp.plotgds(c)

    """
    if not with_elec_connections:
        straight_heater_factory = straight_factory

    cpl = pp.call_if_func(coupler_factory)

    arm_defaults = {
        "L_top": L2,
        "bend_radius": bend_radius,
        "bend90_factory": bend90_factory,
        "straight_heater_factory": straight_heater_factory,
        "straight_factory": straight_factory,
        "with_elec_connections": with_elec_connections,
    }

    arm_top = mzi_arm(L0=L0, **arm_defaults)
    arm_bot = mzi_arm(L0=L0 + DL, **arm_defaults)

    components = {
        "CP1": (cpl, "None"),
        "CP2": (cpl, "mirror_y"),
        "arm_top": (arm_top, "None"),
        "arm_bot": (arm_bot, "mirror_x"),
    }

    connections = [
        # Bottom arm
        ("CP1", "E0", "arm_bot", "W0"),
        ("arm_bot", "E0", "CP2", "E0"),
        # Top arm
        ("CP1", "E1", "arm_top", "W0"),
        ("arm_top", "E0", "CP2", "E0"),
    ]

    if with_elec_connections:
        ports_map = {
            "W0": ("CP1", "W0"),
            "E0": ("CP2", "W0"),
            "E_TOP_0": ("arm_top", "E_0"),
            "E_TOP_1": ("arm_top", "E_1"),
            "E_TOP_2": ("arm_top", "E_2"),
            "E_TOP_3": ("arm_top", "E_3"),
            "E_BOT_0": ("arm_bot", "E_0"),
            "E_BOT_1": ("arm_bot", "E_1"),
            "E_BOT_2": ("arm_bot", "E_2"),
            "E_BOT_3": ("arm_bot", "E_3"),
        }

        component = netlist_to_component(components, connections, ports_map)
        # Need to connect common ground and redefine electrical ports

        ports = component.ports
        y_elec = ports["E_TOP_0"].y
        for ls, le in [
            ("E_BOT_0", "E_BOT_1"),
            ("E_TOP_0", "E_TOP_1"),
            ("E_BOT_2", "E_TOP_2"),
        ]:
            component.add_polygon(line(ports[ls], ports[le]), layer=ports[ls].layer)

        # Add GND
        ("E_BOT_2", "E_TOP_2")
        component.add_port(
            name="GND",
            midpoint=0.5 * (ports["E_BOT_2"].midpoint + ports["E_TOP_2"].midpoint),
            orientation=180,
            width=ports["E_BOT_2"].width,
            layer=ports["E_BOT_2"].layer,
        )

        component.ports["E_TOP_3"].orientation = 0
        component.ports["E_BOT_3"].orientation = 0

        # Remove the eletrical ports that we have just used internally
        for lbl in ["E_BOT_0", "E_BOT_1", "E_TOP_0", "E_TOP_1", "E_BOT_2", "E_TOP_2"]:
            component.ports.pop(lbl)

        # Reroute electrical ports
        _e_ports = select_electrical_ports(component)
        conn, e_ports = route_elec_ports_to_side(_e_ports, side="north", y=y_elec)

        for c in conn:
            component.add(c)

        for p in e_ports:
            component.ports[p.name] = p

        # Create nice electrical port names
        component.ports["HT1"] = component.ports["E_TOP_3"]
        component.ports.pop("E_TOP_3")

        component.ports["HT2"] = component.ports["E_BOT_3"]
        component.ports.pop("E_BOT_3")

    else:
        ports_map = {"W0": ("CP1", "W0"), "E0": ("CP2", "W0")}
        component = netlist_to_component(components, connections, ports_map)

    return component


if __name__ == "__main__":
    import pp

    c = mzi1x2(coupler_factory=mmi1x2, with_elec_connections=False)
    # print(c.ports)
    pp.show(c)
    # print(c.get_settings())



"""
Route optical and electrical waveguides
"""

from pp.routing.add_electrical_pads import add_electrical_pads
from pp.routing.add_electrical_pads_shortest import add_electrical_pads_shortest
from pp.routing.add_electrical_pads_top import add_electrical_pads_top
from pp.routing.add_fiber_array import add_fiber_array
from pp.routing.add_fiber_single import add_fiber_single
from pp.routing.connect import connect_strip, connect_strip_way_points
from pp.routing.connect_bundle import connect_bundle
from pp.routing.connect_bundle import connect_bundle_path_length_match
from pp.routing.connect_bundle import link_electrical_ports
from pp.routing.connect_bundle import link_optical_ports
from pp.routing.manhattan import round_corners, route_manhattan
from pp.routing.repackage import package_optical2x2
from pp.routing.route_fiber_single import route_fiber_single
from pp.routing.route_ports_to_side import route_elec_ports_to_side, route_ports_to_side

__all__ = [
    "add_electrical_pads",
    "add_electrical_pads_shortest",
    "add_electrical_pads_top",
    "add_fiber_array",
    "add_fiber_array",
    "add_fiber_single",
    "connect_bundle",
    "connect_bundle_path_length_match",
    "connect_strip",
    "connect_strip_way_points",
    "link_electrical_ports",
    "link_optical_ports",
    "package_optical2x2",
    "round_corners",
    "route_elec_ports_to_side",
    "route_fiber_single",
    "route_manhattan",
    "route_ports_to_side",
    "route_ports_to_side",
]

if __name__ == "__main__":
    print(__all__)



import pp
from pp.components.rectangle import rectangle_centered
from pp.components.grating_coupler.grating_coupler_tree import grating_coupler_tree
from pp.add_padding import add_padding_to_grid


@pp.autoname
def align_wafer(
    width=10.0,
    spacing=10.0,
    cross_length=80.0,
    layer=pp.LAYER.WG,
    with_tile_excl=True,
    square_corner="bottom_left",
):
    """ returns cross inside a frame to align wafer

    .. plot::
      :include-source:

      import pp

      c = pp.c.align_wafer()
      pp.plotgds(c)

    """
    c = pp.Component()
    cross = pp.c.cross(length=cross_length, width=width, layer=layer)
    c.add_ref(cross)

    b = cross_length / 2 + spacing + width / 2
    w = width

    rh = rectangle_centered(w=2 * b + w, h=w, layer=layer)
    rtop = c.add_ref(rh)
    rbot = c.add_ref(rh)
    rtop.movey(+b)
    rbot.movey(-b)

    rv = rectangle_centered(w=w, h=2 * b, layer=layer)
    rl = c.add_ref(rv)
    rr = c.add_ref(rv)
    rl.movex(-b)
    rr.movex(+b)

    wsq = (cross_length + 2 * spacing) / 4
    square_mark = c << rectangle_centered(w=wsq, h=wsq, layer=layer)
    a = width / 2 + wsq / 2 + spacing

    corner_to_position = {
        "bottom_left": (-a, -a),
        "bottom_right": (a, -a),
        "top_right": (a, a),
        "top_left": (-a, a),
    }

    square_mark.move(corner_to_position[square_corner])

    if with_tile_excl:
        rc_tile_excl = rectangle_centered(
            w=2 * (b + spacing), h=2 * (b + spacing), layer=pp.LAYER.NO_TILE_SI
        )
        c.add_ref(rc_tile_excl)

    return c


@pp.autoname
def add_frame(component, width=10, spacing=10, layer=pp.LAYER.WG):
    """ returns component with a frame around it
    """
    c = pp.Component()
    cref = c.add_ref(component)
    cref.move(-c.size_info.center)
    b = component.size_info.height / 2 + spacing + width / 2
    w = width

    rh = rectangle_centered(w=2 * b + w, h=w, layer=layer)
    rtop = c.add_ref(rh)
    rbot = c.add_ref(rh)
    rtop.movey(+b)
    rbot.movey(-b)

    rv = rectangle_centered(w=w, h=2 * b, layer=layer)
    rl = c.add_ref(rv)
    rr = c.add_ref(rv)
    rl.movex(-b)
    rr.movex(+b)
    c.absorb(cref)

    rc = rectangle_centered(
        w=2 * (b + spacing), h=2 * (b + spacing), layer=pp.LAYER.NO_TILE_SI
    )
    c.add_ref(rc)
    return c


@pp.autoname
def triangle(x, y, layer=1):
    c = pp.Component()
    points = [[x, 0], [0, 0], [0, y]]
    c.add_polygon(points, layer=layer)
    return c


@pp.autoname
def align_cryo_bottom_right(x=60, y=60, layer=1):
    c = align_cryo_top_left()
    cr = c.ref(rotation=180)
    cc = pp.Component()
    cc.add(cr)
    return cc


@pp.autoname
def align_cryo_top_right(x=60, y=60, layer=1):
    c = align_cryo_top_left()
    cr = c.ref(rotation=270)
    cc = pp.Component()
    cc.add(cr)
    return cc


@pp.autoname
def align_cryo_bottom_left(x=60, y=60, layer=1):
    c = align_cryo_top_left()
    cr = c.ref(rotation=90)
    cc = pp.Component()
    cc.add(cr)
    return cc


@pp.autoname
def align_cryo_top_left(x=60, y=60, s=0.2, layer=1):
    c = pp.Component()
    points = [[0, 0], [s, 0], [x - s, y - s], [x - s, y], [0, y]]
    c.add_polygon(points, layer=layer)
    cc = add_frame(component=c)
    cc.name = "align_cryo_top_left"
    return cc


@pp.autoname
def align_tree_top_left(**kwargs):
    c = pp.Component()
    c.name = "grating_coupler_tree_tl"
    gc = grating_coupler_tree(component_name=c.name, **kwargs)
    gc_ref = c.add_ref(gc)
    gc_ref.move(-gc.size_info.center)
    align = align_cryo_top_left()
    c.add_ref(align)
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def align_tree_top_left_with_cross(**kwargs):
    c = pp.Component()
    c.name = "grating_coupler_tree_tl_x"
    gc = grating_coupler_tree(component_name=c.name, **kwargs)
    gc_ref = c.add_ref(gc)
    gc_ref.move(-gc.size_info.center)
    align = align_cryo_top_left()
    c.add_ref(align)
    align2 = align_wafer()
    align2_ref = c.add_ref(align2)
    align2_ref.movex(gc_ref.xmin - align2.size_info.width / 2)
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def align_tree_top_right(**kwargs):
    c = pp.Component()
    c.name = "grating_coupler_tree_tr"
    gc = grating_coupler_tree(component_name=c.name, **kwargs)
    gc_ref = c.add_ref(gc)
    gc_ref.move(-gc.size_info.center)
    align = align_cryo_top_right()
    c.add_ref(align)
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def align_tree_bottom_left(**kwargs):
    c = pp.Component()
    c.name = "grating_coupler_tree_bl"
    gc = grating_coupler_tree(component_name=c.name, **kwargs)
    gc_ref = c.add_ref(gc)
    gc_ref.move(-gc.size_info.center)
    align = align_cryo_bottom_left()
    c.add_ref(align)
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def align_tree_bottom_right(**kwargs):
    c = pp.Component()
    c.name = "grating_coupler_tree_br"
    gc = grating_coupler_tree(component_name=c.name, **kwargs)
    gc_ref = c.add_ref(gc)
    gc_ref.move(-gc.size_info.center)
    align = align_cryo_bottom_right()
    c.add_ref(align)
    c = add_padding_to_grid(c)
    return c


if __name__ == "__main__":
    # c = align_tree_top_left_with_cross()
    c = align_tree_top_left()
    # c = triangle(x=60, y=60)
    # c = align_wafer()
    # c = pp.c.cross(length=80, width=10)
    # c = add_frame(component=c)
    # c = align_cryo_top_right()
    # c = align_cryo_top_left()
    # c = align_cryo_bottom_right()
    pp.show(c)



import pp
from pp import components as pc


@pp.autoname
def test_comb(
    pad_size=(200, 200),
    wire_width=1,
    wire_gap=3,
    comb_layer=0,
    overlap_zigzag_layer=1,
    comb_pad_layer=None,
    comb_gnd_layer=None,
    overlap_pad_layer=None,
):
    """ Superconducting heater device from phidl.geometry

    Args:
        pad_size=(200, 200)
        wire_width=1
        wire_gap=3
        comb_layer=0
        overlap_zigzag_layer=1
        comb_pad_layer=None
        comb_gnd_layer=None
        overlap_pad_layer=None

    """
    CI = pp.Component()

    if comb_pad_layer is None:
        comb_pad_layer = comb_layer
    if comb_gnd_layer is None:
        comb_gnd_layer = comb_layer
    if overlap_pad_layer is None:
        overlap_pad_layer = overlap_zigzag_layer
    wire_spacing = wire_width + wire_gap * 2

    # %% pad overlays
    overlay_padb = CI.add_ref(
        pc.rectangle(
            size=(pad_size[0] * 9 / 10, pad_size[1] * 9 / 10), layer=overlap_pad_layer
        )
    )
    overlay_padl = CI.add_ref(
        pc.rectangle(
            size=(pad_size[0] * 9 / 10, pad_size[1] * 9 / 10), layer=comb_pad_layer
        )
    )
    overlay_padt = CI.add_ref(
        pc.rectangle(
            size=(pad_size[0] * 9 / 10, pad_size[1] * 9 / 10), layer=comb_pad_layer
        )
    )
    overlay_padr = CI.add_ref(
        pc.rectangle(
            size=(pad_size[0] * 9 / 10, pad_size[1] * 9 / 10), layer=comb_gnd_layer
        )
    )

    overlay_padl.xmin = 0
    overlay_padl.ymin = 0
    overlay_padb.ymax = 0
    overlay_padb.xmin = overlay_padl.xmax + pad_size[1] / 5
    overlay_padr.ymin = overlay_padl.ymin
    overlay_padr.xmin = overlay_padb.xmax + pad_size[1] / 5
    overlay_padt.xmin = overlay_padl.xmax + pad_size[1] / 5
    overlay_padt.ymin = overlay_padl.ymax

    # %% pads
    padl = CI.add_ref(pc.rectangle(size=pad_size, layer=comb_layer))
    padt = CI.add_ref(pc.rectangle(size=pad_size, layer=comb_layer))
    padr = CI.add_ref(pc.rectangle(size=pad_size, layer=comb_layer))
    padb = CI.add_ref(pc.rectangle(size=pad_size, layer=overlap_zigzag_layer))
    padl_nub = CI.add_ref(
        pc.rectangle(size=(pad_size[0] / 4, pad_size[1] / 2), layer=comb_layer)
    )
    padr_nub = CI.add_ref(
        pc.rectangle(size=(pad_size[0] / 4, pad_size[1] / 2), layer=comb_layer)
    )

    padl.xmin = overlay_padl.xmin
    padl.center = [padl.center[0], overlay_padl.center[1]]
    padt.ymax = overlay_padt.ymax
    padt.center = [overlay_padt.center[0], padt.center[1]]
    padr.xmax = overlay_padr.xmax
    padr.center = [padr.center[0], overlay_padr.center[1]]
    padb.ymin = overlay_padb.ymin
    padb.center = [overlay_padb.center[0], padb.center[1]]
    padl_nub.xmin = padl.xmax
    padl_nub.center = [padl_nub.center[0], padl.center[1]]
    padr_nub.xmax = padr.xmin
    padr_nub.center = [padr_nub.center[0], padr.center[1]]

    # %% connected zig

    head = CI.add_ref(pc.compass(size=(pad_size[0] / 12, wire_width), layer=comb_layer))
    head.xmin = padl_nub.xmax
    head.ymax = padl_nub.ymax
    connector = CI.add_ref(pc.compass(size=(wire_width, wire_width), layer=comb_layer))
    connector.connect(port="W", destination=head.ports["E"])
    old_port = connector.ports["S"]
    top = True
    obj = connector
    while obj.xmax + pad_size[0] / 12 < padr_nub.xmin:
        # long zig zag rectangle
        obj = CI.add_ref(
            pc.compass(
                size=(pad_size[1] / 2 - 2 * wire_width, wire_width), layer=comb_layer
            )
        )
        obj.connect(port="W", destination=old_port)
        old_port = obj.ports["E"]
        if top:
            # zig zag edge rectangle
            obj = CI.add_ref(
                pc.compass(size=(wire_width, wire_width), layer=comb_layer)
            )
            obj.connect(port="N", destination=old_port)
            top = False
        else:
            # zig zag edge rectangle
            obj = CI.add_ref(
                pc.compass(size=(wire_width, wire_width), layer=comb_layer)
            )
            obj.connect(port="S", destination=old_port)
            top = True
            # comb rectange
            comb = CI.add_ref(
                pc.rectangle(
                    size=(
                        (padt.ymin - head.ymax)
                        + pad_size[1] / 2
                        - (wire_spacing + wire_width) / 2,
                        wire_width,
                    ),
                    layer=comb_layer,
                )
            )
            comb.rotate(90)
            comb.ymax = padt.ymin
            comb.xmax = obj.xmax - (wire_spacing + wire_width) / 2
        old_port = obj.ports["E"]
        obj = CI.add_ref(pc.compass(size=(wire_spacing, wire_width), layer=comb_layer))
        obj.connect(port="W", destination=old_port)
        old_port = obj.ports["E"]
        obj = CI.add_ref(pc.compass(size=(wire_width, wire_width), layer=comb_layer))
        obj.connect(port="W", destination=old_port)
        if top:
            old_port = obj.ports["S"]
        else:
            old_port = obj.ports["N"]
    old_port = obj.ports["E"]
    if padr_nub.xmin - obj.xmax > 0:
        tail = CI.add_ref(
            pc.compass(size=(padr_nub.xmin - obj.xmax, wire_width), layer=comb_layer)
        )
    else:
        tail = CI.add_ref(pc.compass(size=(wire_width, wire_width), layer=comb_layer))
    tail.connect(port="W", destination=old_port)

    # %% disconnected zig

    dhead = CI.add_ref(
        pc.compass(
            size=(padr_nub.ymin - padb.ymax - wire_width, wire_width),
            layer=overlap_zigzag_layer,
        )
    )
    dhead.rotate(90)
    dhead.ymin = padb.ymax
    dhead.xmax = tail.xmin - (wire_spacing + wire_width) / 2
    connector = CI.add_ref(
        pc.compass(size=(wire_width, wire_width), layer=overlap_zigzag_layer)
    )
    connector.connect(port="S", destination=dhead.ports["E"])
    old_port = connector.ports["N"]
    right = True
    obj = connector
    while obj.ymax + wire_spacing + wire_width < head.ymax:
        obj = CI.add_ref(
            pc.compass(size=(wire_spacing, wire_width), layer=overlap_zigzag_layer)
        )
        obj.connect(port="W", destination=old_port)
        old_port = obj.ports["E"]
        if right:
            obj = CI.add_ref(
                pc.compass(size=(wire_width, wire_width), layer=overlap_zigzag_layer)
            )
            obj.connect(port="W", destination=old_port)
            right = False
        else:
            obj = CI.add_ref(
                pc.compass(size=(wire_width, wire_width), layer=overlap_zigzag_layer)
            )
            obj.connect(port="E", destination=old_port)
            right = True
        old_port = obj.ports["N"]
        obj = CI.add_ref(
            pc.compass(
                size=(
                    dhead.xmin - (head.xmax + head.xmin + wire_width) / 2,
                    wire_width,
                ),
                layer=overlap_zigzag_layer,
            )
        )
        obj.connect(port="E", destination=old_port)
        old_port = obj.ports["W"]
        obj = CI.add_ref(
            pc.compass(size=(wire_width, wire_width), layer=overlap_zigzag_layer)
        )
        obj.connect(port="S", destination=old_port)
        if right:
            old_port = obj.ports["W"]
        else:
            old_port = obj.ports["E"]

    return CI


if __name__ == "__main__":
    c = test_comb()
    pp.show(c)



import pp
from pp.mask.merge_json import merge_json
from pp.mask.merge_markdown import merge_markdown
from pp.mask.merge_test_metadata import merge_test_metadata
from pp.mask.write_labels import write_labels


def merge_metadata(gdspath, labels_prefix="opt", label_layer=pp.LAYER.LABEL, **kwargs):
    mdpath = gdspath.with_suffix(".md")
    jsonpath = gdspath.with_suffix(".json")

    build_directory = gdspath.parent.parent
    doe_directory = build_directory / "doe"

    write_labels(gdspath=gdspath, prefix=labels_prefix, label_layer=label_layer)

    merge_json(doe_directory=doe_directory, jsonpath=jsonpath, **kwargs)
    merge_markdown(reports_directory=doe_directory, mdpath=mdpath)
    merge_test_metadata(gdspath, labels_prefix=labels_prefix)


if __name__ == "__main__":

    gdspath = pp.CONFIG["samples_path"] / "mask" / "build" / "mask" / "mask.gds"
    print(gdspath)
    merge_metadata(gdspath)



"""
The choice of font file is configurable from the YAML file
"""
import functools
import klayout.db as pya
import pp
from pp.autoplacer.helpers import load_gds, import_cell


FONT_PATH = pp.CONFIG.get("font_path")


@functools.lru_cache()
def load_alphabet(filepath=FONT_PATH):
    c = load_gds(filepath)
    return {_c.name: _c for _c in c.each_cell()}


def add_text(
    cell, text, position=(0, 0), align_x="center", align_y="top", fontpath=FONT_PATH
):
    """ add text label"""
    text = text.upper()
    alphabet = load_alphabet(filepath=fontpath)
    idbu = 1 / cell.layout().dbu
    si = alphabet["A"].dbbox()
    x, y = position
    w = si.width()
    h = si.height()
    c = cell.layout().create_cell("TEXT_{}".format(text))
    n = len(text)

    if align_x == "center":
        dx = -(n + 0.5) * w / 2
    elif align_x == "right":
        dx = -n * w
    else:
        dx = 0

    if align_y == "top":
        dy = -h
    elif align_y == "center":
        dy = -h / 2
    else:
        dy = 0

    for i, char in enumerate(text):
        _l = import_cell(cell.layout(), alphabet[char])
        _transform = pya.DTrans((i * w + dx) * idbu, dy * idbu)
        label = pya.CellInstArray(_l.cell_index(), _transform)
        c.insert(label)

    cell.insert(pya.CellInstArray(c.cell_index(), pya.DTrans(x, y)))
    return c


def test_alphabet():
    ly = pya.Layout()
    top_cell = ly.create_cell("TOP")
    add_text(top_cell, "HELLO-WORLD_0123456789+_-")
    # add_text(top_cell, "HELLO - WORLD 0123456789 +_-") # Need to add spaces to the font
    top_cell.write("hello_world.gds")


if __name__ == "__main__":
    test_alphabet()



from typing import List, Union

from scipy.special import fresnel
from numpy import pi, sqrt
import numpy as np
from pp.coord2 import Coord2

DEG2RAD = np.pi / 180

# Dict for caching Euler bend point lists
__euler_bend_cache__ = dict()


def euler_bend_points(
    angle_amount: float = 90.0,
    radius: float = 10.0,
    resolution: float = 150.0,
    use_cache: bool = True,
) -> List[Coord2]:
    """ Base euler bend, no transformation, emerging from the origin."""
    # Check if we've calculated this already
    key = (angle_amount, radius, resolution)
    if key in __euler_bend_cache__ and use_cache:
        return __euler_bend_cache__[key]

    if angle_amount < 0:
        raise ValueError("angle_amount should be positive. Got {}".format(angle_amount))
    # End angle
    eth = angle_amount * DEG2RAD

    # If bend is trivial, return a trivial shape
    if eth == 0.0:
        return [Coord2(0, 0)]

    # Curve min radius
    R = radius

    # Total displaced angle
    th = eth / 2.0

    # Total length of curve
    Ltot = 4 * R * th

    # Compute curve ##
    a = sqrt(R ** 2.0 * np.abs(th))
    sq2pi = sqrt(2.0 * pi)

    # Function for computing curve coords
    (fasin, facos) = fresnel(sqrt(2.0 / pi) * R * th / a)

    def _xy(s):
        if th == 0:
            return Coord2(0.0, 0.0)
        elif s <= Ltot / 2:
            (fsin, fcos) = fresnel(s / (sq2pi * a))
            X = sq2pi * a * fcos
            Y = sq2pi * a * fsin
        else:
            (fsin, fcos) = fresnel((Ltot - s) / (sq2pi * a))
            X = (
                sq2pi
                * a
                * (
                    facos
                    + np.cos(2 * th) * (facos - fcos)
                    + np.sin(2 * th) * (fasin - fsin)
                )
            )
            Y = (
                sq2pi
                * a
                * (
                    fasin
                    - np.cos(2 * th) * (fasin - fsin)
                    + np.sin(2 * th) * (facos - fcos)
                )
            )
        return Coord2(X, Y)

    # Parametric step size
    step = Ltot / int(th * resolution)

    # Generate points
    points = []
    for i in range(0, int(round(Ltot / step)) + 1):
        points += [_xy(i * step)]

    # Cache calculated points
    if use_cache:
        __euler_bend_cache__[(angle_amount, radius, resolution)] = points

    return points


def euler_end_pt(
    start_point=(0.0, 0.0), radius=10.0, input_angle=0.0, angle_amount=90.0
):
    """Gives the end point of a simple Euler bend as a Coord2"""

    th = abs(angle_amount) * DEG2RAD / 2.0
    R = radius
    clockwise = bool(angle_amount < 0)

    (fsin, fcos) = fresnel(sqrt(2 * th / pi))

    a = 2 * sqrt(2 * pi * th) * (np.cos(th) * fcos + np.sin(th) * fsin)
    r = a * R
    X = r * np.cos(th)
    Y = r * np.sin(th)

    if clockwise:
        Y *= -1

    pt = Coord2(X, Y) + start_point
    pt.rotate(rotation_center=start_point, rotation=input_angle)
    return pt


def euler_length(radius: Union[int, float] = 10.0, angle_amount: int = 90.0) -> float:
    th = abs(angle_amount) * DEG2RAD / 2
    return 4 * radius * th



import numpy as np
from numpy import cos, sin, pi, sqrt

import pp
from pp.component import Component
from typing import Tuple


@pp.autoname
def ellipse(
    radii: Tuple[float, float] = (10.0, 5.0),
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = pp.LAYER.WG,
) -> Component:
    """Generate an ellipse geometry.

    Args:
        radii: (tuple) Semimajor and semiminor axis lengths of the ellipse.
        angle_resolution: (float) Resolution of the curve of the ring (# of degrees per point).
        layer: (int, array-like[2], or set) Specific layer(s) to put polygon geometry on.

    The orientation of the ellipse is determined by the order of the radii variables;
    if the first element is larger, the ellipse will be horizontal and if the second
    element is larger, the ellipse will be vertical.

    .. plot::
      :include-source:

      import pp

      c = pp.c.ellipse(radii=(10, 5), angle_resolution=2.5, layer=0)
      pp.plotgds(c)

    """

    D = pp.Component()
    a = radii[0]
    b = radii[1]
    t = np.linspace(0, 360, int(360 / angle_resolution) + 1) * pi / 180
    r = a * b / (sqrt((b * cos(t)) ** 2 + (a * sin(t)) ** 2))
    xpts = r * cos(t)
    ypts = r * sin(t)
    D.add_polygon(points=(xpts, ypts), layer=layer)
    return D


if __name__ == "__main__":
    c = ellipse()
    pp.show(c)



import os
import itertools
import klayout.db as pya
import pp.autoplacer.functions as ap
from pp.autoplacer.library import Library
from pp.autoplacer.auto_placer import AutoPlacer


class ChipArray(AutoPlacer):
    """ An array of chiplets with dicing lanes
    """

    def __init__(
        self,
        name,
        mask_width,
        mask_height,
        cols,
        rows,
        lib,
        spacing=25000,
        lane_width=50000,
        align=None,
    ):
        super(ChipArray, self).__init__(name, mask_width, mask_height)
        self.lib = lib
        self.rows = rows
        self.cols = cols
        self.spacing = spacing
        self.lane_width = lane_width
        self.align = align

        # Infer chip width and height
        self.chip_width = (mask_width - (cols - 1) * spacing) / cols
        self.chip_height = (mask_height - (rows - 1) * spacing) / rows

        self.make_chips()
        # self.make_dicing_lanes()

    def make_chips(self):
        """ Make all the chips """
        # Get the aligntree
        if self.align:
            aligntrees = self.lib.get(self.align)
            self.aligntree = aligntrees[0]

        # Make all the chips
        self.chips = []
        for row, col in itertools.product(
            list(range(self.rows)), list(range(self.cols))
        ):
            name = "{}{}".format(row, col)
            chip = AutoPlacer(name, self.chip_width, self.chip_height)
            if self.align:
                for corner in ap.CORNERS:
                    chip.pack_auto(self.aligntree, corner)

            chip.draw_boundary()
            chip.row, chip.col = row, col
            self.chips.append(chip)

    def make_dicing_lanes(self):
        """ Make the dicing lanes """
        container = self.create_cell("DicingLanes")
        instance = pya.CellInstArray(container.cell_index(), pya.Trans(0, 0))
        self.cell(self.name).insert(instance)
        lw = self.lane_width / 2

        for dicing_layer in ap.DICING_LAYERS:
            layer = self.layer(dicing_layer[0], dicing_layer[1])

            for row in range(1, self.rows):
                y = row * (self.chip_height + self.spacing) - self.spacing / 2
                box = pya.Box(0, y - lw, self.max_width, y + lw)
                container.shapes(layer).insert(box)

            for col in range(1, self.cols):
                x = col * (self.chip_width + self.spacing) - self.spacing / 2
                for row in range(self.rows):
                    y1 = row * (self.chip_height + self.spacing)
                    y2 = (row + 1) * (self.chip_height + self.spacing) - self.spacing
                    box = pya.Box(x - lw, y1, x + lw, y2)
                    container.shapes(layer).insert(box)

            # on the corners, line has half the width
            lw = self.lane_width / 4

            for row in [0, self.rows]:
                y = row * (self.chip_height + self.spacing) - self.spacing / 2
                box = pya.Box(0, y - lw, self.max_width, y + lw)
                container.shapes(layer).insert(box)

            for col in [0, self.cols]:
                x = col * (self.chip_width + self.spacing) - self.spacing / 2
                for row in range(self.rows):
                    y1 = row * (self.chip_height + self.spacing)
                    y2 = (row + 1) * (self.chip_height + self.spacing) - self.spacing
                    box = pya.Box(x - lw, y1, x + lw, y2)
                    container.shapes(layer).insert(box)

    def write(self, *args, **kwargs):
        """ Write to disk. We pack the chips at the last minute. """
        self.draw_boundary(ap.DEVREC_LAYER)
        self.draw_boundary(ap.FLOORPLAN_LAYER)
        for chip in self.chips:
            x = chip.col * (self.chip_width + self.spacing)
            y = chip.row * (self.chip_height + self.spacing)
            self.pack_manual(chip.top_cell(), x, y)
        super(ChipArray, self).write(*args, **kwargs)

    def write_chips(self, name=None, path=None):
        if name is None:
            name = self.name
        if path is None:
            path = os.path.join("build", "mask")
        filename = os.path.join(path, name)
        for chip in self.chips:
            chip.write(filename + "_" + chip.name + ".gds")


if __name__ == "__main__":
    lib = Library()
    mask = ChipArray("chip_array", 25e6, 25e6, 3, 4, lib)
    mask.pack_grid(lib.pop("align"))
    mask.pack_grid(lib.pop(".*"))
    mask.write("build/chip_array.gds")



import numpy as np
import pp
from pp import Port
from pp.routing.connect_bundle_from_waypoints import connect_bundle_waypoints


@pp.autoname
def test_connect_bundle_waypoints():
    return test_connect_bundle_waypointsA()


@pp.autoname
def test_connect_bundle_waypointsA():

    xs1 = np.arange(10) * 5 - 500.0

    N = xs1.size
    ys2 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 75]) + 500.0

    ports1 = [Port("A_{}".format(i), (xs1[i], 0), 0.5, 90) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (0, ys2[i]), 0.5, 180) for i in range(N)]

    top_cell = pp.Component()
    p0 = ports1[0].position + (22, 0)
    way_points = [
        p0,
        p0 + (0, 100),
        p0 + (200, 100),
        p0 + (200, -200),
        p0 + (0, -200),
        p0 + (0, -350),
        p0 + (400, -350),
        (p0[0] + 400, ports2[-1].y),
        ports2[-1].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsB():

    ys1 = np.array([0, 5, 10, 15, 30, 40, 50, 60]) + 0.0
    ys2 = np.array([0, 10, 20, 30, 70, 90, 110, 120]) + 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (500, ys2[i]), 0.5, 180) for i in range(N)]

    p0 = ports1[0].position + (0, 22.5)

    top_cell = pp.Component()
    way_points = [
        p0,
        p0 + (200, 0),
        p0 + (200, -200),
        p0 + (400, -200),
        (p0[0] + 400, ports2[0].y),
        ports2[0].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsC():

    ys1 = np.array([0, 5, 10, 15, 20, 60, 70, 80, 120, 125])
    ys2 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 65]) - 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (600, ys2[i]), 0.5, 180) for i in range(N)]

    top_cell = pp.Component()
    way_points = [
        ports1[0].position,
        ports1[0].position + (200, 0),
        ports1[0].position + (200, -200),
        ports1[0].position + (400, -200),
        (ports1[0].x + 400, ports2[0].y),
        ports2[0].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsD():

    ys1 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 75]) + 100.0
    ys2 = np.array([0, -5, -10, -20, -25, -30, -40, -55, -60, -75]) + 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (0, ys2[i]), 0.5, 0) for i in range(N)]

    def _mean_y(ports):
        return np.mean([p.y for p in ports])

    yc1 = _mean_y(ports1)
    yc2 = _mean_y(ports2)

    top_cell = pp.Component()
    way_points = [(0, yc1), (200, yc1), (200, yc2), (0, yc2)]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


if __name__ == "__main__":
    cell = test_connect_bundle_waypointsD()
    pp.show(cell)



""" adapted from phidl.Geometry
"""

import rectpack
import numpy as np
from pp.component import Component
from numpy import ndarray
from typing import Any, Dict, List, Tuple


def _pack_single_bin(
    rect_dict: Dict[int, Tuple[int, int]],
    aspect_ratio: Tuple[int, int],
    max_size: ndarray,
    sort_by_area: bool,
    density: float,
    precision: float,
    verbose: bool,
) -> Tuple[Dict[int, Tuple[int, int, int, int]], Dict[Any, Any]]:
    """ Takes a `rect_dict` argument of the form {id:(w,h)} and tries to
    pack it into a bin as small as possible with aspect ratio `aspect_ratio`
    Will iteratively grow the bin size until everything fits or the bin size
    reaches `max_size`.

    Returns: a dictionary of of the packed rectangles in the form {id:(x,y,w,h)}, and a dictionary of remaining unpacked rects
    """

    # Compute total area and use it for an initial estimate of the bin size
    total_area = 0
    for r in rect_dict.values():
        total_area += r[0] * r[1]
    aspect_ratio = np.asarray(aspect_ratio) / np.linalg.norm(aspect_ratio)  # Normalize

    # Setup variables
    box_size = np.asarray(aspect_ratio * np.sqrt(total_area), dtype=np.float64)
    box_size = np.clip(box_size, None, max_size)
    if sort_by_area:
        rp_sort = rectpack.SORT_AREA
    else:
        rp_sort = rectpack.SORT_NONE

    # Repeatedly run the rectangle-packing algorithm with increasingly larger
    # areas until everything fits or we've reached the maximum size
    while True:
        # Create the pack object
        rect_packer = rectpack.newPacker(
            mode=rectpack.PackingMode.Offline,
            pack_algo=rectpack.MaxRectsBlsf,
            sort_algo=rp_sort,
            bin_algo=rectpack.PackingBin.BBF,
            rotation=False,
        )

        # Add each rectangle to the pack, create a single bin, and pack
        for rid, r in rect_dict.items():
            rect_packer.add_rect(width=r[0], height=r[1], rid=rid)
        rect_packer.add_bin(width=box_size[0], height=box_size[1])
        rect_packer.pack()

        # Adjust the box size for next time
        box_size *= density  # Increase area to try to fit
        box_size = np.clip(box_size, None, max_size)
        if verbose:
            print(
                "Trying to pack in bin size (%0.2f, %0.2f)"
                % tuple(box_size * precision)
            )

        # Quit the loop if we've packed all the rectangles or reached the max size
        if len(rect_packer.rect_list()) == len(rect_dict):
            if verbose:
                print("Success!")
            break
        elif all(box_size >= max_size):
            if verbose:
                print("Reached max_size, creating an additional bin")
            break

    # Separate packed from unpacked rectangles, make dicts of form {id:(x,y,w,h)}
    packed_rect_dict = {r[-1]: r[:-1] for r in rect_packer[0].rect_list()}
    unpacked_rect_dict = {}
    for k, v in rect_dict.items():
        if k not in packed_rect_dict:
            unpacked_rect_dict[k] = v

    return (packed_rect_dict, unpacked_rect_dict)


def pack(
    D_list: List[Component],
    spacing: int = 10,
    aspect_ratio: Tuple[int, int] = (1, 1),
    max_size: Tuple[None, None] = (None, None),
    sort_by_area: bool = True,
    density: float = 1.1,
    precision: float = 1e-2,
    verbose: bool = False,
) -> List[Component]:
    """ takes a list of components and returns

    Args:
        D_list: Must be a list or tuple of Components
        spacing: Minimum distance between adjacent shapes
        aspect_ratio: (width, height) ratio of the rectangular bin
        max_size: Limits the size into which the shapes will be packed
        density:  Values closer to 1 pack tighter but require more computation
        sort_by_area (Boolean): Pre-sorts the shapes by area
        verbose: False
    """

    if density < 1.01:
        raise ValueError(
            "pack() was given a `density` argument that is"
            + " too small.  The density argument must be >= 1.01"
        )

    # Santize max_size variable
    max_size = [np.inf if v is None else v for v in max_size]
    max_size = np.asarray(max_size, dtype=np.float64)  # In case it's integers
    max_size = max_size / precision

    # Convert Components to rectangles
    rect_dict = {}
    for n, D in enumerate(D_list):
        w, h = (D.size + spacing) / precision
        w, h = int(w), int(h)
        if (w > max_size[0]) or (h > max_size[1]):
            raise ValueError(
                "pack() failed because one of the objects "
                + "in `D_list` is has an x or y dimension larger than `max_size` and "
                + "so cannot be packed"
            )
        rect_dict[n] = (w, h)

    packed_list = []
    while len(rect_dict) > 0:
        (packed_rect_dict, rect_dict) = _pack_single_bin(
            rect_dict,
            aspect_ratio=aspect_ratio,
            max_size=max_size,
            sort_by_area=sort_by_area,
            density=density,
            precision=precision,
            verbose=verbose,
        )
        packed_list.append(packed_rect_dict)

    D_packed_list = []
    for rect_dict in packed_list:
        D_packed = Component()
        for n, rect in rect_dict.items():
            x, y, w, h = rect
            xcenter = x + w / 2 + spacing / 2
            ycenter = y + h / 2 + spacing / 2
            d = D_packed.add_ref(D_list[n])
            d.center = (xcenter * precision, ycenter * precision)
        D_packed_list.append(D_packed)

    return D_packed_list


def _demo():
    import pp
    import phidl.geometry as pg

    D_list = [pg.ellipse(radii=np.random.rand(2) * n + 2) for n in range(50)]
    D_list += [pg.rectangle(size=np.random.rand(2) * n + 2) for n in range(50)]

    D_packed_list = pack(
        D_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        max_size=(None, None),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
        verbose=False,
    )
    D = D_packed_list[0]  # Only one bin was created, so we plot that
    pp.show(D)  # show it in klayout


if __name__ == "__main__":
    import pp
    import phidl.geometry as pg

    spacing = 1
    ellipses = pack(
        [pg.ellipse(radii=np.random.rand(2) * n + 2) for n in range(50)],
        spacing=spacing,
    )[0]
    ellipses.name = "ellipses"
    rectangles = pack(
        [pg.rectangle(size=np.random.rand(2) * n + 2) for n in range(50)],
        spacing=spacing,
    )[0]
    rectangles.name = "rectangles"
    p = pack([ellipses, rectangles])
    pp.show(p[0])



import os
from pp.load_component import load_component


def hash_geometry(gdspath):
    """ the Geometrical hash includes only layers and polygons
    """
    assert os.path.isfile(gdspath), "!{} not found or not a valid GDS path".format(
        gdspath
    )
    name = gdspath.stem
    path = gdspath.parent
    c = load_component(name, path)
    return c.hash_geometry()


def same_hash(gdspath1, gdspath2):
    """ returns True if hash is the same between 2 saved gds files
    ignores timestamps
    """
    return hash_geometry(gdspath1) == hash_geometry(gdspath2)
    # return gdspy.gdsii_hash(gdspath1) == gdspy.gdsii_hash(gdspath2)


if __name__ == "__main__":
    import pp

    gdspath = pp.write_component_type(
        "mmi1x2", width_mmi=4, overwrite=True, path_directory=pp.CONFIG["gdslib"]
    )
    print(hash_geometry(gdspath))

    gdspath = pp.write_component_type(
        "mmi1x2", width_mmi=5, overwrite=True, path_directory=pp.CONFIG["gdslib_test"]
    )
    print(hash_geometry(gdspath))



from typing import Callable, List, Tuple
from pp.name import autoname
from pp.layers import LAYER
from pp.components.compass import compass
from pp.component import Component

WIRE_WIDTH = 10.0


@autoname
def pad(
    width: int = 100, height: int = 100, layer: Tuple[int, int] = LAYER.M3
) -> Component:
    """rectangular pad with 4 ports (N, S, E, W)

    Args:
        width: pad width
        height: pad height
        layer: pad layer


    .. plot::
      :include-source:

      import pp

      c = pp.c.pad(width=100, height=100, layer=pp.LAYER.M3)
      pp.plotgds(c)

    """
    c = Component()
    _c = compass(size=(width, height), layer=layer).ref()
    c.add(_c)
    c.absorb(_c)
    c.ports = _c.ports
    return c


@autoname
def pad_array(
    pad: Callable = pad,
    spacing: Tuple[int, int] = (150.0, 0.0),
    n: int = 6,
    port_list: List[str] = ["N"],
    width: float = 100.0,
    height: float = 100.0,
    layer: Tuple[int, int] = LAYER.M3,
) -> Component:
    """array of rectangular pads

    Args:
        pad: pad element
        spacing: (x, y) spacing
        n: number of pads
        port_list: list of port orientations (N, S, W, E) per pad
        width: pad width
        height: pad height
        layer: pad layer

    .. plot::
      :include-source:

      import pp

      c = pp.c.pad_array(pad=pp.c.pad, spacing=(150, 0), n=6, port_list=["N"])
      pp.plotgds(c)

    """
    c = Component()
    pad = pad(width=width, height=height, layer=layer) if callable(pad) else pad

    for i in range(n):
        p = c << pad
        p.x = i * spacing[0]
        for port_name in port_list:
            port_name_new = "{}{}".format(port_name, i)
            c.add_port(port=p.ports[port_name], name=port_name_new)

    return c


if __name__ == "__main__":
    import pp

    c = pad()
    print(c.ports)
    # c = pad(width=10, height=10)
    # print(c.ports.keys())
    # print(c.settings['spacing'])
    # c = pad_array()
    pp.show(c)



""" ensures gdslibrary exists
"""

import pp


def test_gdslib_exists():
    assert pp.CONFIG["gdslib"].exists()


if __name__ == "__main__":
    test_gdslib_exists()
    # print(pp.CONFIG['gdslib'])



from typing import Callable, Optional
import pp

from pp.component import Component
from pp.port import deco_rename_ports, rename_ports_by_orientation
from pp.components import bend_circular
from pp.components import waveguide
from pp.components import mmi1x2


@deco_rename_ports
@pp.autoname
def mzi(
    L0: float = 1.0,
    DL: float = 0.1,
    L2: float = 20.1,
    bend_radius: float = 10.0,
    bend90: Callable = bend_circular,
    waveguide: Callable = waveguide,
    waveguide_vertical: Optional[Callable] = None,
    waveguide_horizontal: Optional[Callable] = None,
    coupler: Callable = mmi1x2,
    combiner: Optional[Callable] = None,
) -> Component:
    """Mzi

    Args:
        L0: vertical length for both and top arms
        DL: bottom arm extra length, (delta_length = 2*DL)
        L2: L_top horizontal length
        bend_radius: 10.0
        bend90: bend_circular
        waveguide_vertical: waveguide
        coupler: coupler
        combiner: coupler

    .. code::

               __L2__
              |      |
              L0     L0r
              |      |
     coupler==|      |==combiner
              |      |
              L0     L0r
              |      |
              DL     DL
              |      |
              |__L2__|


    .. plot::
      :include-source:

      import pp

      c = pp.c.mzi(L0=0.1, DL=0, L2=10)
      pp.plotgds(c)

    """
    c = pp.Component()
    coupler = pp.call_if_func(coupler)
    if combiner:
        combiner = pp.call_if_func(combiner)
    else:
        combiner = coupler

    waveguide_vertical = waveguide_vertical or waveguide
    waveguide_horizontal = waveguide_horizontal or waveguide
    b90 = bend90(radius=bend_radius) if callable(bend90) else bend90
    l0 = waveguide_vertical(length=L0)

    coupler = rename_ports_by_orientation(coupler)
    combiner = rename_ports_by_orientation(combiner)

    y1l = coupler.ports["E0"].y
    y1r = combiner.ports["E0"].y

    y2l = coupler.ports["E1"].y
    y2r = combiner.ports["E1"].y

    dl = abs(y2l - y1l)  # splitter ports distance
    dr = abs(y2r - y1r)  # combiner ports distance
    delta_length = dl - dr
    assert delta_length + L0 > 0, (
        f"input and output couplers height  offset (delta_length)  {delta_length} +"
        f" {L0} >0"
    )

    l0r = waveguide_vertical(length=L0 + delta_length / 2)
    l1 = waveguide_vertical(length=DL)
    l2 = waveguide_horizontal(length=L2)

    cin = c << coupler
    cout = c << combiner

    # top arm
    blt = c << b90
    bltl = c << b90
    bltr = c << b90
    blmr = c << b90  # bend left medium right

    l0tl = c << l0
    l2t = c << l2
    l0tr = c << l0r

    blt.connect(port="W0", destination=cin.ports["E1"])
    l0tl.connect(port="W0", destination=blt.ports["N0"])
    bltl.connect(port="N0", destination=l0tl.ports["E0"])
    l2t.connect(port="W0", destination=bltl.ports["W0"])
    bltr.connect(port="N0", destination=l2t.ports["E0"])
    l0tr.connect(port="W0", destination=bltr.ports["W0"])
    blmr.connect(port="W0", destination=l0tr.ports["E0"])
    cout.connect(port="E0", destination=blmr.ports["N0"])

    # bot arm
    blb = c << b90
    l0bl = c << l0
    l1l = c << l1
    blbl = c << b90
    l2t = c << l2
    brbr = c << b90
    l1r = c << l1
    l0br = c << l0r
    blbmrb = c << b90  # bend left medium right bottom

    blb.connect(port="N0", destination=cin.ports["E0"])
    l0bl.connect(port="W0", destination=blb.ports["W0"])
    l1l.connect(port="W0", destination=l0bl.ports["E0"])
    blbl.connect(port="W0", destination=l1l.ports["E0"])
    l2t.connect(port="W0", destination=blbl.ports["N0"])
    brbr.connect(port="W0", destination=l2t.ports["E0"])

    l1r.connect(port="W0", destination=brbr.ports["N0"])
    l0br.connect(port="W0", destination=l1r.ports["E0"])
    blbmrb.connect(port="N0", destination=l0br.ports["E0"])
    blbmrb.connect(port="W0", destination=cout.ports["E1"])  # just for netlist

    # west ports
    for port_name, port in cin.ports.items():
        if port.angle == 180:
            c.add_port(name=port_name, port=port)

    # east ports
    i = 0
    for port_name, port in cout.ports.items():
        if port.angle == 0:
            c.add_port(name=f"E{i}", port=port)
            i += 1

    return c


if __name__ == "__main__":
    DL = 116.8 / 2
    # print(DL)
    c = mzi(DL=DL, pins=True)
    # print(c.ports["E0"].midpoint[1])
    # c.plot_netlist()
    print(c.ports.keys())
    print(c.ports["E0"].midpoint)
    pp.show(c)
    pp.qp(c)
    # print(c.get_settings())



import json
from pp.name import get_component_name
from pp.components import component_type2factory
from pp.write_component import write_component
from pp.config import CONFIG
from pp.doe import get_settings_list
from pp.routing.add_fiber_array import add_fiber_array_te, add_fiber_array_tm

name2function = dict(
    add_fiber_array_te=add_fiber_array_te, add_fiber_array_tm=add_fiber_array_tm
)


def write_doe_metadata(
    doe_name,
    cell_names,
    list_settings,
    doe_settings=None,
    doe_metadata_path=CONFIG["doe_directory"],
    **kwargs,
):
    """ writes DOE metadata (markdown report, JSON dict)

    Args:
        doe_name
        cell_names: list of klayout <Cell> or pp <Component>
        list_settings: list of settings for the cell
        doe_settings: test and data_analysis_protocol
        cell_settings: list of cell settings

    """

    doe_metadata_path.mkdir(parents=True, exist_ok=True)
    report_path = doe_metadata_path / (doe_name + ".md")
    doe_settings = doe_settings or {}
    json_path = report_path.with_suffix(".json")

    d = dict(
        type="doe",
        name=doe_name,
        cells={cell_name: {"name": cell_name} for cell_name in cell_names},
        settings=list_settings,
        doe_settings=doe_settings,
        **kwargs,
    )

    with open(json_path, "w+") as fw:
        fw.write(json.dumps(d, indent=2))

    with open(report_path, "w+") as fw:

        def w(line=""):
            fw.write(line + "\n")

        w("# {}".format(doe_name))
        w("- Number of devices: {}".format(len(list_settings)))
        w("- Settings")

        if len(kwargs) > 0:
            w(json.dumps(doe_settings, indent=2))

        # w(json.dumps(list_settings))
        if list_settings:

            w()
            EOL = ""

            # Convert table fields to strings
            head = [
                [str(field) for field in fields.keys()] for fields in list_settings
            ][0]
            list_data_str = [
                [str(field) for field in fields.values()] for fields in list_settings
            ]
            N = len(head)

            field_sizes = [0] * N

            # compute the max number of character in each field
            for header, fields in zip(head, list_data_str):
                field_sizes = [
                    max(field_sizes[i], len(fields[i]), len(head[i])) for i in range(N)
                ]

            field_sizes = [n + 2 for n in field_sizes]

            # Line formatting from fields
            def fmt_line(fields):
                fmt_fields = [
                    " " + fields[i] + " " * (field_sizes[i] - len(fields[i]) - 1)
                    for i in range(N)
                ]
                return "|".join(fmt_fields) + EOL

            def table_head_sep():
                return "|".join(["-" * n for n in field_sizes]) + EOL

            w(fmt_line(head))
            w(table_head_sep())
            for fields in list_data_str[0:]:
                w(fmt_line(fields))

            w()
            w("Cells: \n")
            for cell_name in cell_names:
                w(f"- {cell_name}")

            w()


def write_doe(
    component_type,
    doe_name=None,
    do_permutations=True,
    list_settings=None,
    doe_settings=None,
    path=CONFIG["build_directory"],
    doe_metadata_path=CONFIG["doe_directory"],
    functions=None,
    name2function=name2function,
    **kwargs,
):
    """ writes each device GDS, together with metadata for each device:
    Returns a list of gdspaths

    - .gds geometry for each component
    - .json metadata for each component
    - .ports if any for each component
    - report.md for DOE
    - doe_name.json with DOE metadata

    pp.write_component_doe("mmi1x2", width_mmi=[5, 10], length_mmi=9)

    Args:
        component_type: component_name_or_function
        doe_name: autoname by default
        do_permutations: builds all permutations between the varying parameters
        list_settings: you can pass a list of settings or the variations in the kwargs
        doe_settings: shared settings for a DOE
        path: to store build artifacts
        functions: list of function names to apply to DOE
        name2function: function names to functions dict
        **kwargs: Doe default settings or variations
    """
    if hasattr(component_type, "__call__"):
        component_type = component_type.__name__

    doe_name = doe_name or component_type
    functions = functions or []
    list_settings = list_settings or get_settings_list(
        do_permutations=do_permutations, **kwargs
    )

    assert isinstance(component_type, str), "{} not recognized".format(component_type)

    path.mkdir(parents=True, exist_ok=True)

    doe_gds_paths = []
    cell_names = []
    cell_settings = []

    for settings in list_settings:
        # print(settings)
        component_name = get_component_name(component_type, **settings)
        component_factory = component_type2factory[component_type]
        component = component_factory(name=component_name, **settings)
        component.function_name = component_factory.__name__
        if "test" in kwargs:
            component.test_protocol = kwargs.get("test")
        if "analysis" in kwargs:
            component.data_analysis_protocol = kwargs.get("analysis")
        for f in functions:
            component = name2function[f](component)

        cell_names.append(component.name)
        cell_settings.append(settings)
        gdspath = path / (component.name + ".gds")
        doe_gds_paths += [gdspath]
        write_component(component, gdspath)

    """ write DOE metadata (report + JSON) """
    write_doe_metadata(
        doe_name=doe_name,
        cell_names=cell_names,
        list_settings=list_settings,
        doe_settings=doe_settings,
        cell_settings=kwargs,
        doe_metadata_path=doe_metadata_path,
    )

    return doe_gds_paths


def get_markdown_table(do_permutations=True, **kwargs):
    """ returns the markdown table for a parameter sweep
    """
    list_settings = get_settings_list(do_permutations=do_permutations, **kwargs)
    # Convert table fields to strings
    head = [[str(field) for field in fields.keys()] for fields in list_settings][0]
    list_data_str = [
        [str(field) for field in fields.values()] for fields in list_settings
    ]
    N = len(head)

    field_sizes = [0] * N

    # compute the max number of character in each field
    for header, fields in zip(head, list_data_str):
        field_sizes = [
            max(field_sizes[i], len(fields[i]), len(head[i])) for i in range(N)
        ]

    field_sizes = [n + 2 for n in field_sizes]

    # Line formatting from fields
    def fmt_line(fields):
        fmt_fields = [
            " " + fields[i] + " " * (field_sizes[i] - len(fields[i]) - 1)
            for i in range(N)
        ]
        return "|".join(fmt_fields)

    def table_head_sep():
        return "|".join(["-" * n for n in field_sizes])

    t = []
    t.append(fmt_line(head))
    t.append(table_head_sep())
    for fields in list_data_str[0:]:
        t.append(fmt_line(fields))

    return t


def test_write_doe():
    paths = write_doe(
        "mmi1x2",
        width_mmi=[5, 10],
        length_mmi=[20, 30],
        do_permutations=True,
        functions=["add_fiber_array_tm"],
        doe_settings=dict(test="optical_tm"),
    )
    assert len(paths) == 4

    # print(type(paths[0]))
    # pp.show(paths[0])

    paths = write_doe(
        "mmi1x2", width_mmi=[5, 10], length_mmi=[20, 30], do_permutations=False
    )
    assert len(paths) == 2


if __name__ == "__main__":
    # print(get_markdown_table(width_mmi=[5, 6]))
    test_write_doe()
    # paths = write_doe(
    #     "mmi1x2", width_mmi=[5, 10], length_mmi=[20, 30], do_permutations=False
    # )
    # print(paths)
    # gdspaths = test_write_doe()



""" write DOE from YAML file """

import sys
import importlib

from pp.config import CONFIG
from pp.doe import load_does
from pp.write_doe import write_doe


def import_custom_doe_factories():
    """ Find if we have custom DOEs on this config.
    Make them available in component_type2factory
    """

    sys.path += [CONFIG["mask_root_directory"]]
    if CONFIG["custom_components"]:
        try:
            importlib.import_module(CONFIG["custom_components"])
        except Exception:
            pass


def write_doe_from_yaml(filepath):
    """ Loads DOE settings from yaml file and writes GDS into build_directory

    Args:
        filepath: YAML file describing DOE

    Returns:
        gdspaths: list

    For each DOE save:

    - GDS
    - json metadata
    - ports CSV
    - markdown report, with DOE settings
    """
    does = load_does(filepath)

    gdspaths = []
    for doe_name, doe in does.items():
        # print(doe_name)
        # print(doe.get("settings"))
        # print(doe.get("do_permutations"))
        # print(doe)
        # print(list(doe.keys()))
        # print(type(doe.get('settings')))
        # assert type(doe.get('settings'))
        d = write_doe(
            component_type=doe.get("component"),
            doe_name=doe_name,
            do_permutations=doe.get("do_permutations", True),
            list_settings=doe.get("settings"),
            description=doe.get("description"),
            analysis=doe.get("analysis"),
            test=doe.get("test"),
            functions=doe.get("functions"),
        )
        gdspaths.append(d)
    return gdspaths


def test_write_doe_from_yaml():
    does_path = CONFIG["samples_path"] / "mask" / "does.yml"
    gdspaths = write_doe_from_yaml(does_path)
    print(gdspaths)


if __name__ == "__main__":
    test_write_doe_from_yaml()



import gdspy
import numpy as np

from phidl.geometry import (
    Device,
    DeviceReference,
    _merge_floating_point_errors,
    _parse_layer,
    Polygon,
    _offset_polygons_parallel,
)
import pp


def offset(
    elements,
    distance=0.1,
    join_first=True,
    precision=1e-4,
    num_divisions=[1, 1],
    join="miter",
    tolerance=2,
    max_points=4000,
    layer=0,
):
    """ returns an element containing all polygons with an offset
    from phidl geometry
    """
    if type(elements) is not list:
        elements = [elements]
    polygons_to_offset = []
    for e in elements:
        if isinstance(e, (Device, DeviceReference)):
            polygons_to_offset += e.get_polygons(by_spec=False)
        elif isinstance(e, (Polygon, gdspy.Polygon)):
            polygons_to_offset.append(e)
    if len(polygons_to_offset) == 0:
        return pp.Component("offset")
    polygons_to_offset = _merge_floating_point_errors(
        polygons_to_offset, tol=precision / 1000
    )
    gds_layer, gds_datatype = _parse_layer(layer)
    if all(np.array(num_divisions) == np.array([1, 1])):
        p = gdspy.offset(
            polygons_to_offset,
            distance=distance,
            join=join,
            tolerance=tolerance,
            precision=precision,
            join_first=join_first,
            max_points=max_points,
            layer=gds_layer,
            datatype=gds_datatype,
        )
    else:
        p = _offset_polygons_parallel(
            polygons_to_offset,
            distance=distance,
            num_divisions=num_divisions,
            join_first=join_first,
            precision=precision,
            join=join,
            tolerance=tolerance,
        )

    D = pp.Component("offset")
    polygons = D.add_polygon(p, layer=layer)
    [
        polygon.fracture(max_points=max_points, precision=precision)
        for polygon in polygons
    ]
    return D


if __name__ == "__main__":
    c = pp.c.ring()
    co = offset(c)
    pp.show(co)



import numpy as np

import pp
from pp.components.bezier import bezier
from pp.component import Component
from typing import List, Tuple


@pp.autoname
def bend_s(
    width: float = 0.5,
    height: float = 2.0,
    length: float = 10.0,
    layer: Tuple[int, int] = pp.LAYER.WG,
    nb_points: int = 99,
    layers_cladding: List[Tuple[int, int]] = [pp.LAYER.WGCLAD],
    cladding_offset: float = 3.0,
) -> Component:
    """ S bend
    Based on bezier curve

    Args:
        width
        height: in y direction
        length: in x direction
        layer: gds number
        nb_points: number of points

    .. plot::
      :include-source:

      import pp

      c = pp.c.bend_s(width=0.5, height=20)
      pp.plotgds(c)

    """
    l, h = length, height
    c = bezier(
        width=width,
        control_points=[(0, 0), (l / 2, 0), (l / 2, h), (l, h)],
        t=np.linspace(0, 1, nb_points),
        layer=layer,
    )
    c.add_port(name="W0", port=c.ports.pop("0"))
    c.add_port(name="E0", port=c.ports.pop("1"))

    y = cladding_offset
    points = [
        [c.xmin, c.ymin - y],
        [c.xmax, c.ymin - y],
        [c.xmax, c.ymax + y],
        [c.xmin, c.ymax + y],
    ]
    for layer in layers_cladding:
        c.add_polygon(points, layer=layer)

    # c.ports["W0"] = c.ports.pop("0")
    # c.ports["E0"] = c.ports.pop("1")
    return c


@pp.port.deco_rename_ports
@pp.autoname
def bend_s_biased(width=0.5, height=2, length=10, layer=pp.LAYER.WG, nb_points=99):
    l, h = length, height
    return bezier(
        width=pp.bias.width(width),
        control_points=[(0, 0), (l / 2, 0), (l / 2, h), (l, h)],
        t=np.linspace(0, 1, nb_points),
        layer=layer,
    )


def _demo():
    c = bend_s()
    pp.write_gds(c)
    return c


if __name__ == "__main__":
    c = bend_s(pins=True)
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    pp.show(c)



import pp


def test_autoname():
    wg1 = pp.c.waveguide(length=10, width=0.5)
    wg2 = pp.c.waveguide(width=0.5, length=10)
    assert (
        wg1.name == wg2.name
    ), f"{wg1} and {wg2} waveguides have the same settings and should have the same name"
    return wg1, wg2


if __name__ == "__main__":
    wg1, wg2 = test_autoname()
    print(wg1)
    print(wg2)



import pp


def test_mzi():

    netlist = """
    instances:
        CP1:
          component: mmi1x2
          settings:
              width_mmi: 4.5
              length_mmi: 10
        CP2:
            component: mmi1x2
            settings:
                width_mmi: 4.5
                length_mmi: 5
        arm_top:
            component: mzi_arm
        arm_bot:
            component: mzi_arm

    placements:
        arm_bot:
            mirror: [0,0,0,10]
    ports:
        W0: CP1,W0
        E0: CP2,W0

    connections:
        arm_bot,W0: CP1,E0
        arm_top,W0: CP1,E1
        CP2,E0: arm_bot,E0
        CP2,E1: arm_top,E0
    """

    return pp.component_from_yaml(netlist)


if __name__ == "__main__":
    c = test_mzi()
    pp.show(c)
    pp.plotgds(c)



import csv
import numpy as np


def load_csv(csv_path):
    """ loads csv and returs a dict
    the key for each column is taken from the first row
    """
    dict_data = {}
    with open(csv_path, "r", encoding="utf-8-sig") as csvfile:
        csv_reader = csv.reader(csvfile, delimiter=",")
        fields = next(csv_reader)
        dict_data = {field: [] for field in fields}
        for row in csv_reader:
            for i, x in enumerate(row):
                field = fields[i]
                dict_data[field].append(float(x))

    dict_data = {k: np.array(v) for k, v in list(dict_data.items())}
    return dict_data


def load_csv_pandas(csv_path):
    import pandas as pd

    return pd.read_csv(csv_path)


if __name__ == "__main__":
    from pp.config import CONFIG

    csv_path = CONFIG["components_path"] / "csv_data" / "taper_strip_0p5_3_36.csv"

    d = load_csv_pandas(csv_path)
    print(d)



""" from a gdsfactory component write Sparameters from an FDTD Lumerical simulation
"""

import json

from collections import namedtuple
import numpy as np
import pp
from pp.layers import layer2material, layer2nm
from pp.config import materials


def write(
    component,
    session=None,
    run=True,
    overwrite=False,
    dirpath=pp.CONFIG["sp"],
    height_nm=220,
    **settings,
):
    """
    writes Sparameters from a gdsfactory component using Lumerical FDTD

    Args:
        component: gdsfactory Component
        session: you can pass a session=lumapi.FDTD() for debugging
        run: True-> runs Lumerical , False -> only draws simulation
        overwrite: run even if simulation results already exists
        dirpath: where to store the simulations
        height_nm: height
        layer2nm: dict of {(1, 0): 220}
        layer2material: dict of {(1, 0): "si"}
        remove_layers: list of tuples (layers to remove)
        background_material: for the background
        port_width: port width (m)
        port_height: port height (m)
        port_extension_um: port extension (um)
        mesh_accuracy: 2 (1: coarse, 2: fine, 3: superfine)
        zmargin: for the FDTD region 1e-6 (m)
        ymargin: for the FDTD region 2e-6 (m)
        wavelength_start: 1.2e-6 (m)
        wavelength_stop: 1.6e-6 (m)
        wavelength_points: 500

    Return:
        results: dict(wavelength_nm, S11, S12 ...) after simulation, or if simulation exists and returns the Sparameters directly
    """
    if hasattr(component, "simulation_settings"):
        settings.update(component.simulation_settings)
    sim_settings = s = dict(
        layer2nm=layer2nm,
        layer2material=layer2material,
        remove_layers=[pp.LAYER.WGCLAD],
        background_material="sio2",
        port_width=3e-6,
        port_height=1.5e-6,
        port_extension_um=1,
        mesh_accuracy=2,
        zmargin=1e-6,
        ymargin=2e-6,
        wavelength_start=1.2e-6,
        wavelength_stop=1.6e-6,
        wavelength_points=500,
    )
    for setting in settings.keys():
        assert (
            setting in s
        ), f"`{setting}` is not a valid setting ({list(settings.keys())})"
    s.update(**settings)
    ss = namedtuple("sim_settings", s.keys())(*s.values())

    assert ss.port_width < 5e-6
    assert ss.port_height < 5e-6
    assert ss.zmargin < 5e-6
    assert ss.ymargin < 5e-6

    ports = component.ports

    component.remove_layers(ss.remove_layers)
    component._bb_valid = False

    c = pp.extend_ports(component, length=ss.port_extension_um)
    gdspath = pp.write_gds(c)

    filepath = component.get_sparameters_path(dirpath=dirpath, height_nm=height_nm)
    filepath_json = filepath.with_suffix(".json")
    filepath_sim_settings = filepath.with_suffix(".settings.json")
    filepath_fsp = filepath.with_suffix(".fsp")

    if run and filepath_json.exists() and not overwrite:
        return json.loads(open(filepath_json).read())

    if not run and session is None:
        print(
            """
you need to pass `run=True` flag to run the simulation
To debug, you can create a lumerical FDTD session and pass it to the simulator

```
import lumapi
s = lumapi.FDTD()

import pp
c = pp.c.waveguide() # or whatever you want to simulate
pp.sp.write(component=c, run=False, session=s)
```
"""
        )

    pe = ss.port_extension_um * 1e-6 / 2
    x_min = c.xmin * 1e-6 + pe
    x_max = c.xmax * 1e-6 - pe

    y_min = c.ymin * 1e-6 - ss.ymargin
    y_max = c.ymax * 1e-6 + ss.ymargin

    port_orientations = [p.orientation for p in ports.values()]
    if 90 in port_orientations and len(ports) > 2:
        y_max = c.ymax * 1e-6 - pe
        x_max = c.xmax * 1e-6

    elif 90 in port_orientations:
        y_max = c.ymax * 1e-6 - pe
        x_max = c.xmax * 1e-6 + ss.ymargin

    z = 0
    z_span = 2 * ss.zmargin + max(ss.layer2nm.values()) * 1e-9

    import lumapi

    s = session or lumapi.FDTD(hide=False)
    s.newproject()
    s.selectall()
    s.deleteall()
    s.addrect(
        x_min=x_min,
        x_max=x_max,
        y_min=y_min,
        y_max=y_max,
        z=z,
        z_span=z_span,
        index=1.5,
        name="clad",
    )

    material = ss.background_material
    if material not in materials:
        raise ValueError(f"{material} not in {list(materials.keys())}")
    material = materials[material]
    s.setnamed("clad", "material", material)

    s.addfdtd(
        dimension="3D",
        x_min=x_min,
        x_max=x_max,
        y_min=y_min,
        y_max=y_max,
        z=z,
        z_span=z_span,
        mesh_accuracy=ss.mesh_accuracy,
        use_early_shutoff=True,
    )

    layers = component.get_layers()
    for layer, nm in ss.layer2nm.items():
        if layer not in layers:
            continue
        assert layer in ss.layer2material, f"{layer} not in {ss.layer2material.keys()}"

        material = ss.layer2material[layer]
        if material not in materials:
            raise ValueError(f"{material} not in {list(materials.keys())}")
        material = materials[material]

        s.gdsimport(str(gdspath), c.name, f"{layer[0]}:{layer[1]}")
        silicon = f"GDS_LAYER_{layer[0]}:{layer[1]}"
        s.setnamed(silicon, "z span", nm * 1e-9)
        s.setnamed(silicon, "material", material)

    for i, port in enumerate(ports.values()):
        s.addport()
        p = f"FDTD::ports::port {i+1}"
        s.setnamed(p, "x", port.x * 1e-6)
        s.setnamed(p, "y", port.y * 1e-6)
        s.setnamed(p, "z span", ss.port_height)

        deg = int(port.orientation)
        # assert port.orientation in [0, 90, 180, 270], f"{port.orientation} needs to be [0, 90, 180, 270]"
        if -45 <= deg <= 45:
            direction = "Backward"
            injection_axis = "x-axis"
            dxp = 0
            dyp = ss.port_width
        elif 45 < deg < 90 + 45:
            direction = "Backward"
            injection_axis = "y-axis"
            dxp = ss.port_width
            dyp = 0
        elif 90 + 45 < deg < 180 + 45:
            direction = "Forward"
            injection_axis = "x-axis"
            dxp = 0
            dyp = ss.port_width
        elif 180 + 45 < deg < -45:
            direction = "Forward"
            injection_axis = "y-axis"
            dxp = ss.port_width
            dyp = 0

        else:
            raise ValueError(
                f"port {port.name} with orientation {port.orientation} is not a valid"
                " number "
            )

        s.setnamed(p, "direction", direction)
        s.setnamed(p, "injection axis", injection_axis)
        s.setnamed(p, "y span", dyp)
        s.setnamed(p, "x span", dxp)
        # s.setnamed(p, "theta", deg)
        s.setnamed(p, "name", port.name)

    s.setglobalsource("wavelength start", ss.wavelength_start)
    s.setglobalsource("wavelength stop", ss.wavelength_stop)
    s.setnamed("FDTD::ports", "monitor frequency points", ss.wavelength_points)

    if run:
        s.save(str(filepath_fsp))
        s.deletesweep("s-parameter sweep")

        s.addsweep(3)
        s.setsweep("s-parameter sweep", "Excite all ports", 0)
        s.setsweep("S sweep", "auto symmetry", True)
        s.runsweep("s-parameter sweep")

        # collect results
        # S_matrix = s.getsweepresult("s-parameter sweep", "S matrix")
        sp = s.getsweepresult("s-parameter sweep", "S parameters")

        # export S-parameter data to file named s_params.dat to be loaded in INTERCONNECT
        s.exportsweep("s-parameter sweep", str(filepath))
        print(f"wrote sparameters to {filepath}")

        keys = [key for key in sp.keys() if key.startswith("S")]
        ra = {f"{key}a": list(np.unwrap(np.angle(sp[key].flatten()))) for key in keys}
        rm = {f"{key}m": list(np.abs(sp[key].flatten())) for key in keys}

        results = {"wavelength_nm": list(sp["lambda"].flatten() * 1e9)}
        results.update(ra)
        results.update(rm)
        with open(filepath_json, "w") as f:
            json.dump(results, f)

        with open(filepath_sim_settings, "w") as f:
            s = sim_settings
            s["layer2nm"] = [f"{k[0]}_{k[1]}_{v}" for k, v in s["layer2nm"].items()]
            s["layer2material"] = [
                f"{k[0]}_{k[1]}_{v}" for k, v in s["layer2material"].items()
            ]
            json.dump(s, f)

        return results


def write_coupler_ring():
    [
        write(
            pp.c.coupler_ring(
                wg_width=wg_width, length_x=length_x, bend_radius=bend_radius, gap=gap
            )
            for wg_width in [0.5]
            for length_x in [0.1, 1, 2, 3, 4]
            for gap in [0.15, 0.2]
            for bend_radius in [5, 10]
        )
    ]


if __name__ == "__main__":
    c = pp.c.coupler_ring(length_x=3)
    r = write(component=c)
    print(r)
    # print(r.keys())
    # print(c.ports.keys())



from pp.port import deco_rename_ports
from pp.container import container
from pp.component import Component


@container
@deco_rename_ports
def rotate(component: Component, angle: int = 90) -> Component:
    """ returns rotated component
    """
    c = Component(
        settings=component.get_settings(),
        test_protocol=component.test_protocol,
        data_analysis_protocol=component.data_analysis_protocol,
    )
    cr = c.add_ref(component)
    cr.rotate(angle)
    c.ports = cr.ports
    c.name = component.name + "_r"
    return c


if __name__ == "__main__":
    import pp

    c = pp.c.mmi1x2()
    cr = rotate(c)
    pp.show(cr)
    print(cr.ports)



import pp


@pp.autoname
def sims_pad(width=500, height=500, pad=100, layer=1):
    c = pp.Component()
    w = width
    h = height
    points = [
        [-w / 2.0, -h / 2.0],
        [-w / 2.0, h / 2],
        [w / 2, h / 2],
        [w / 2, -h / 2.0],
    ]
    c.add_polygon(points, layer=layer)
    w = width + 2 * pad
    h = height + 2 * pad
    points = [
        [-w / 2.0, -h / 2.0],
        [-w / 2.0, h / 2],
        [w / 2, h / 2],
        [w / 2, -h / 2.0],
    ]
    c.add_polygon(points, layer=pp.LAYER.PADDING)
    return c


if __name__ == "__main__":
    c = sims_pad()
    pp.show(c)



import shutil
import pytest

from pp import CONFIG


@pytest.fixture(autouse=True)
def cleandir(request):
    build_folder = CONFIG["build_directory"]
    module_path = CONFIG["module_path"]

    if build_folder.exists() and "noautofixt" not in request.keywords:
        shutil.rmtree(build_folder)

    for build_folder in module_path.glob("**/build"):
        shutil.rmtree(build_folder)



from pp.autoplacer.yaml_placer import update_dicts_recurse


def test1():
    target_dict = {
        "placer": {"x0": 1, "y0": 1},
        "nested": {"nested1": {"nested2": {"toto": "toto"}}},
    }

    default_dict = {
        "placer": {"x0": 2, "y0": 2, "x": 0, "y": 200},
        "name": {"suffix": "TEG1"},
        "nested": {"nested1": {"nested2": {"foo": "bar"}}},
    }

    new_dict = update_dicts_recurse(target_dict.copy(), default_dict.copy())
    print(new_dict)


if __name__ == "__main__":
    test1()



import numpy as np

import picwriter.components as pc

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component
from typing import Callable, Tuple


@pp.autoname
def coupler_full(
    length: float = 40.0,
    gap: float = 0.5,
    dw: float = 0.1,
    angle: float = np.pi / 6,
    parity: int = 1,
    port: Tuple[int, int] = (0, 0),
    direction: str = "EAST",
    waveguide_template: Callable = wg_strip,
    **kwargs
) -> Component:
    """ Adiabatic Full Coupler.  Design based on asymmetric adiabatic full coupler designs, such as the one reported in 'Integrated Optic Adiabatic Devices on Silicon' by Y. Shani, et al (IEEE Journal of Quantum Electronics, Vol. 27, No. 3 March 1991).

    In this design, Region I is the first half of the input S-bend waveguide where the input waveguides widths taper by +dw and -dw, Region II is the second half of the S-bend waveguide with constant, unbalanced widths, Region III is the coupling region where the waveguides from unbalanced widths to balanced widths to reverse polarity unbalanced widths, Region IV is the fixed width waveguide that curves away from the coupling region, and Region V is the final curve where the waveguides taper back to the regular width specified in the waveguide template.

    Args:
       length (float): Length of the coupling region.
       gap (float): Distance between the two waveguides.
       dw (float): Change in waveguide width.  Top arm tapers to the waveguide width - dw, bottom taper to width - dw.
       angle (float): Angle in radians (between 0 and pi/2) at which the waveguide bends towards the coupling region.  Default=pi/6.
       parity (integer -1 or 1): If -1, mirror-flips the structure so that the input port is actually the *bottom* port.  Default = 1.
       port (tuple): Cartesian coordinate of the input port (AT TOP if parity=1, AT BOTTOM if parity=-1).  Defaults to (0,0).
       direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians).  Defaults to 'EAST'.
       waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate object

    Other Parameters:
       wg_width: 0.5
       wg_layer: pp.LAYER.WG[0]
       wg_datatype: pp.LAYER.WG[1]
       clad_layer: pp.LAYER.WGCLAD[0]
       clad_datatype: pp.LAYER.WGCLAD[1]
       bend_radius: 10
       cladding_offset: 3


    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler_full(length=40, gap=0.2, dw=0.1)
      pp.plotgds(c)

    """

    c = pc.FullCoupler(
        pp.call_if_func(waveguide_template, **kwargs),
        length=length,
        gap=gap,
        dw=dw,
        angle=angle,
        parity=parity,
        port=port,
        direction=direction,
    )

    return picwriter2component(c)


if __name__ == "__main__":
    import pp

    c = coupler_full(length=40, gap=0.2, dw=0.1)
    pp.show(c)



""" write Component from YAML file """

import io
from typing import Union, IO, Any
import pathlib
from omegaconf import OmegaConf

from pp.component import Component
from pp.components import component_type2factory


sample = """
mmi_long:
  component: mmi1x2
  settings:
    width_mmi: 4.5
    length_mmi: 10
  properties:
    x : 100
    y : 100
mmi_short:
  component: mmi1x2
  settings:
    width_mmi: 4.5
    length_mmi: 5
  properties:
    x : 120
    y : 100
"""


def read_yaml(yaml: Union[str, pathlib.Path, IO[Any]]) -> Component:
    """ Loads Components settings from yaml file and writes the GDS into build_directory

    Args:
        yaml: YAML IO describing DOE

    Returns:
        Component
    """
    c = Component()

    yaml = io.StringIO(yaml) if isinstance(yaml, str) and "\n" in yaml else yaml
    conf = OmegaConf.load(yaml)

    for component_name in conf:
        component_conf = conf[component_name]
        component_type = component_conf["component"]
        component_settings = component_conf["settings"]
        ci = component_type2factory[component_type](**component_settings)
        component_properties = component_conf["properties"]
        for k, v in component_properties.items():
            setattr(ci, k, v)
        ci.name = component_name
        c << ci
    return c


def test_read_yaml():
    pass


if __name__ == "__main__":
    import pp

    c = read_yaml(sample)
    pp.show(c)



""" bends with grating couplers inside the spiral
maybe: need to add grating coupler loopback as well
"""

import numpy as np
import pp
from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular180
from pp.components.euler.bend_euler import bend_euler90, bend_euler180
from pp.components.waveguide import waveguide
from pp.routing import round_corners
from pp.config import TAPER_LENGTH
from pp.component import Component
from typing import Callable, Optional, Tuple


def get_bend_port_distances(bend: Component) -> Tuple[float, float]:
    p0, p1 = bend.ports.values()
    return abs(p0.x - p1.x), abs(p0.y - p1.y)


@pp.autoname
def spiral_inner_io(
    N: int = 6,
    x_straight_inner_right: float = 150.0,
    x_straight_inner_left: float = 150.0,
    y_straight_inner_top: float = 50.0,
    y_straight_inner_bottom: float = 10.0,
    grating_spacing: float = 127.0,
    dx: float = 3.0,
    dy: float = 3.0,
    bend90_function: Callable = bend_circular,
    bend180_function: Callable = bend_circular180,
    bend_radius: float = 50.0,
    wg_width: float = 0.5,
    wg_width_grating_coupler: float = 0.5,
    straight_factory: Callable = waveguide,
    taper: Optional[Callable] = None,
    length: Optional[float] = None,
) -> Component:
    """

    Args:
        N: number of loops
        x_straight_inner_right:
        x_straight_inner_left:
        y_straight_inner_top:
        y_straight_inner_bottom:
        grating_spacing:
        dx: center to center x-spacing
        dy: center to center y-spacing
        bend90_function
        bend180_function
        bend_radius
        wg_width
        straight_factory
        taper:
        length: cm

    .. plot::
      :include-source:

      import pp
      from pp.components.spiral_inner_io import spiral_inner_io

      c = spiral_inner_io()
      pp.plotgds(c)
    """
    if length:
        if bend180_function == bend_circular180:
            y_straight_inner_top = get_straight_length(
                length_cm=length,
                spiral_function=spiral_inner_io,
                N=N,
                x_straight_inner_right=x_straight_inner_right,
                x_straight_inner_left=x_straight_inner_left,
                y_straight_inner_top=y_straight_inner_top,
                y_straight_inner_bottom=y_straight_inner_bottom,
                grating_spacing=grating_spacing,
                dx=dx,
                dy=dy,
                straight_factory=waveguide,
                bend90_function=bend_euler90,
                bend180_function=bend_euler180,
                wg_width=wg_width,
            )
        else:
            y_straight_inner_top = get_straight_length(
                length_cm=length,
                spiral_function=spiral_inner_io_euler,
                N=N,
                x_straight_inner_right=x_straight_inner_right,
                x_straight_inner_left=x_straight_inner_left,
                y_straight_inner_top=y_straight_inner_top,
                y_straight_inner_bottom=y_straight_inner_bottom,
                grating_spacing=grating_spacing,
                dx=dx,
                dy=dy,
                wg_width=wg_width,
            )

    _bend180 = pp.call_if_func(bend180_function, radius=bend_radius, width=wg_width)
    _bend90 = pp.call_if_func(bend90_function, radius=bend_radius, width=wg_width)

    rx, ry = get_bend_port_distances(_bend90)
    _, rx180 = get_bend_port_distances(_bend180)  # rx180, second arg since we rotate

    component = pp.Component()
    # gc_port_lbl = "W0"
    # gc1 = _gc.ref(port_id=gc_port_lbl, position=(0, 0), rotation=-90)
    # gc2 = _gc.ref(port_id=gc_port_lbl, position=(grating_spacing, 0), rotation=-90)
    # component.add([gc1, gc2])

    p1 = pp.Port(
        name="S0",
        midpoint=(0, y_straight_inner_top),
        orientation=270,
        width=wg_width,
        layer=pp.LAYER.WG,
    )
    p2 = pp.Port(
        name="S1",
        midpoint=(grating_spacing, y_straight_inner_top),
        orientation=270,
        width=wg_width,
        layer=pp.LAYER.WG,
    )
    taper = pp.c.taper(
        width1=wg_width_grating_coupler,
        width2=_bend180.ports["W0"].width,
        length=TAPER_LENGTH + y_straight_inner_top - 15 - 35,
    )
    taper_ref1 = component.add_ref(taper)
    taper_ref1.connect("2", p1)

    taper_ref2 = component.add_ref(taper)
    taper_ref2.connect("2", p2)

    component.absorb(taper_ref1)
    component.absorb(taper_ref2)

    component.add_port(name="S0", port=taper_ref1.ports["1"])
    component.add_port(name="S1", port=taper_ref2.ports["1"])

    # Create manhattan path going from west grating to westest port of bend 180
    _pt = np.array(p1.position)
    pts_w = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i + 1) * dy
        x2 = grating_spacing + 2 * rx + x_straight_inner_right + (2 * i + 1) * dx
        y3 = -y_straight_inner_bottom - ry - (2 * i + 3) * dy
        x4 = -x_straight_inner_left - (2 * i + 1) * dx
        if i == N - 1:
            x4 = x4 - rx180 + dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_w += [_pt1, _pt2, _pt3, _pt4, _pt5]

    route_ref_w = round_corners(
        pts_w, bend90=_bend90, straight_factory=straight_factory, taper=taper
    )
    component.add(route_ref_w)
    component.absorb(route_ref_w)

    # Add loop back
    bend180_ref = _bend180.ref(
        port_id="W1", position=route_ref_w.ports["output"], rotation=90
    )
    component.add(bend180_ref)
    component.absorb(bend180_ref)

    # Create manhattan path going from east grating to eastest port of bend 180
    _pt = np.array(p2.position)
    pts_e = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i) * dy
        x2 = grating_spacing + 2 * rx + x_straight_inner_right + 2 * i * dx
        y3 = -y_straight_inner_bottom - ry - (2 * i + 2) * dy
        x4 = -x_straight_inner_left - (2 * i) * dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_e += [_pt1, _pt2, _pt3, _pt4, _pt5]

    route_ref_e = round_corners(
        pts_e, bend90=_bend90, straight_factory=straight_factory, taper=taper
    )
    component.add(route_ref_e)
    component.absorb(route_ref_e)

    length = (
        route_ref_e.info["length"]
        + route_ref_w.info["length"]
        + bend180_ref.info["length"]
    )
    component.length = length + 2 * y_straight_inner_top
    # print(length * 1e-6 * 1e2, "cm")

    return component


@pp.autoname
def spiral_inner_io_euler(
    bend90_function: Callable = bend_euler90,
    bend180_function: Callable = bend_euler180,
    **kwargs
) -> Component:
    """

    .. plot::
      :include-source:

      import pp

      c = pp.c.spiral_inner_io_euler()
      pp.plotgds(c)
    """

    return spiral_inner_io(
        bend90_function=bend90_function, bend180_function=bend180_function, **kwargs
    )


@pp.autoname
def spirals_nested(bend_radius=100):
    component = pp.Component()
    c = spiral_inner_io(
        N=42,
        y_straight_inner_top=10.0,
        y_straight_inner_bottom=5700.0,
        x_straight_inner_right=2000.0,
        x_straight_inner_left=20.0,
        bend_radius=bend_radius,
    )

    c1 = spiral_inner_io(
        N=42,
        y_straight_inner_top=10.0,
        y_straight_inner_bottom=10.0,
        x_straight_inner_right=0.0,
        x_straight_inner_left=120.0,
        bend_radius=bend_radius,
    )

    c2 = spiral_inner_io(
        N=42,
        y_straight_inner_top=10.0,
        y_straight_inner_bottom=2000.0,
        x_straight_inner_right=0.0,
        x_straight_inner_left=120.0,
        bend_radius=bend_radius,
    )

    # for _c in [c, c1, c2]:
    #     print(_c.info["length"])

    component.add(c.ref(position=(0, 0)))
    component.add(c1.ref(position=(1150, -850)))
    component.add(c2.ref(position=(1150, -2850)))
    return component


@pp.autoname
def mini_block_mockup():
    component = pp.Component()

    c = spirals_nested()
    x0 = c.size_info.west
    y0 = c.size_info.south
    margin = 50.0
    component.add(c.ref(position=(-x0 + margin, -y0 + margin)))
    component.add(c.ref(position=(4500 - x0 + margin, -y0 + margin)))

    return component


@pp.autoname
def reticle_mockup():
    from pp.layers import LAYER

    dx = 9000.0
    dy = 8000.0
    a0 = 50.0
    component = pp.Component()
    for x in [0, dx, 2 * dx, 3 * dx]:
        a = x + a0
        b = x - a0

        component.add_polygon(
            [(a, 0), (b, 0), (b, 3 * dy), (a, 3 * dy)], LAYER.FLOORPLAN
        )

    for y in [0, dy, 2 * dy, 3 * dy]:
        a = y + a0
        b = y - a0
        component.add_polygon(
            [(0, a), (0, b), (3 * dx, b), (3 * dx, a)], LAYER.FLOORPLAN
        )

    c = mini_block_mockup()

    for i, j in [(0, 0), (2, 2), (0, 2), (2, 0), (1, 1)]:
        component.add(c.ref(position=(i * dx, j * dy)))

    return component


def get_straight_length(length_cm, spiral_function, **kwargs):
    """ returns y_spiral to achieve a particular spiral length """
    y0 = 50
    y1 = 400
    kwargs.update({"y_straight_inner_top": y0})
    s0 = spiral_function(**kwargs)
    kwargs.update({"y_straight_inner_top": y1})
    s1 = spiral_function(**kwargs)
    p = np.polyfit(
        np.array([y0, y1]), 1e-6 * 1e2 * np.array([s0.length, s1.length]), deg=1
    )
    return (length_cm - p[1]) / p[0]


# @pp.autoname
# def spiral_inner_io_with_gratings(
#     spiral=spiral_inner_io, grating_coupler=pp.c.grating_coupler_elliptical_te, **kwargs
# ):
#     spiral = pp.call_if_func(spiral, **kwargs)
#     grating_coupler = pp.call_if_func(grating_coupler)

#     return add_gratings_and_loop_back(spiral, grating_coupler=grating_coupler)


if __name__ == "__main__":
    from pp.add_termination import add_gratings_and_loop_back

    c = spiral_inner_io()
    # c = spiral_inner_io_euler()
    # c = spiral_inner_io_euler(length=2, wg_width=0.4)
    # c = spiral_inner_io_euler(length=6, wg_width=0.4)
    print(c.name)
    print(c.settings)
    cc = add_gratings_and_loop_back(c)
    pp.show(c)

    # c = spiral_inner_io_euler(wg_width=1)
    # from pp.routing import add_fiber_array
    # c = spiral_inner_io_euler(length=4, wg_width=1)
    # cc = pp.routing.add_fiber_array(c)
    # print(c.length)
    # print(get_straight_length(2, spiral_inner_io_euler))
    # print(get_straight_length(4, spiral_inner_io_euler))
    # print(get_straight_length(6, spiral_inner_io_euler))
    # c = spiral_inner_io()
    # c = spiral_inner_io_euler(y_straight_inner_top=-11)
    # c = spiral_inner_io_euler(bend_radius=20, wg_width=0.2)
    # c = spiral_inner_io_euler(bend_radius=20, wg_width=0.2, y_straight_inner_top=200)
    # c = reticle_mockup()
    # c = spiral_inner_io()
    # c = spiral_inner_io(bend_radius=20, wg_width=0.2)
    # c = spirals_nested()
    pp.show(c)



from numpy import pi
import pp
from pp.components import taper
from pp.components import bend_circular
from pp.components import waveguide
from pp.routing.manhattan import round_corners
from pp.config import WG_EXPANDED_WIDTH, TAPER_LENGTH
from pp.component import Component
from typing import Callable


@pp.autoname
def delay_snake(
    total_length: float = 160000.0,
    L0: float = 2350.0,
    n: int = 5,
    taper: Callable = taper,
    bend_factory: Callable = bend_circular,
    bend_radius: float = 10.0,
    straight_factory: Callable = waveguide,
    wg_width: float = 0.5,
) -> Component:
    """ Snake input facing west
    Snake output facing east

    Args:
        total_length:
        L0:
        n: number of loops
        taper:
        bend_factory
        bend_radius
        straight_factory
        wg_width

    .. code::

       | L0 |    L2        |

            ->-------------|
                           | pi * radius
       |-------------------|
       |
       |------------------->

       |        DL         |

    .. plot::
      :include-source:

      import pp

      c = pp.c.delay_snake(L0=5, total_length=1600, n=2)
      pp.plotgds(c)

    """
    epsilon = 0.1
    R = bend_radius
    bend90 = bend_factory(radius=R, width=wg_width)
    DL = (total_length + L0 - n * (pi * R + epsilon)) / (2 * n + 1)
    L2 = DL - L0
    assert (
        L2 > 0
    ), "Snake is too short: either reduce L0, increase the total length,\
    or decrease n"

    y = 0
    path = [(0, y), (L2, y)]
    for i in range(n):
        y -= 2 * R + epsilon
        path += [(L2, y), (-L0, y)]
        y -= 2 * R + epsilon
        path += [(-L0, y), (L2, y)]

    path = [(round(_x, 3), round(_y, 3)) for _x, _y in path]

    component = pp.Component()
    if taper:
        if callable(taper):
            _taper = taper(
                width1=wg_width, width2=WG_EXPANDED_WIDTH, length=TAPER_LENGTH
            )
        else:
            _taper = taper
    snake = round_corners(path, bend90, straight_factory, taper=_taper)
    component.add(snake)
    component.ports = snake.ports

    pp.port.auto_rename_ports(component)
    return component


if __name__ == "__main__":
    c = delay_snake()
    pp.show(c)



"""
This is a convenience function for cascading components. Usecase, is composite
waveguides such as phase modulators, where we need to keep track of multiple tapers,
doped sections, undopped, heaters etc...

The idea is to associate one symbol per type of section.
A section is uniquely defined by the component, its selected input and its selected output.

The mapping between symbols and components is supplied by a dictionnary.
The actual chain of components is supplied by a string or a list
"""

import pp
from pp.components import bend_circular
from pp.components.waveguide import waveguide
from pp.components.waveguide_heater import waveguide_heater
from pp.components.component_sequence import component_sequence


@pp.autoname
def test_cutback_heater():
    # Define subcomponents
    bend_radius = 10.0
    bend180 = bend_circular(radius=bend_radius, start_angle=-90, theta=180)
    wg = waveguide(length=5.0)
    wg_heater = waveguide_heater(length=20.0)

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (bend180, "W0", "W1"),
        "B": (bend180, "W1", "W0"),
        "H": (wg_heater, "W0", "E0"),
        "-": (wg, "W0", "E0"),
    }

    # Generate a sequence
    # This is simply a chain of characters. Each of them represents a component
    # with a given input and and a given output

    sequence = "AB-H-H-H-H-BA"
    component = component_sequence(sequence, string_to_device_in_out_ports)
    assert component
    return component


if __name__ == "__main__":
    c = test_cutback_heater()
    pp.show(c)



import numpy as np
import pp
from pp.components.bezier import bezier
from pp.container import container


@container
def package_optical2x2(component, port_spacing=20.0, bend_length=None):
    """returns component with port_spacing"""

    component = pp.call_if_func(component)
    component.y = 0

    if bend_length is None:
        bend_length = port_spacing
    dx = bend_length
    dy = port_spacing / 2.0

    p_w0 = component.ports["W0"].midpoint
    p_e0 = component.ports["E0"].midpoint
    p_w1 = component.ports["W1"].midpoint
    p_e1 = component.ports["E1"].midpoint

    c = pp.Component(f"{component.name}_{int(port_spacing)}")
    c.add_ref(component)

    t = np.linspace(0, 1, 101)

    y0 = p_e1[1]
    control_points = [(0, 0), (dx / 2, 0), (dx / 2, dy - y0), (dx, dy - y0)]

    bezier_bend_t = bezier(
        control_points=control_points, t=t, start_angle=0, end_angle=0
    )

    b_tr = bezier_bend_t.ref(port_id="0", position=p_e1)
    b_br = bezier_bend_t.ref(port_id="0", position=p_e0, v_mirror=True)
    b_tl = bezier_bend_t.ref(port_id="0", position=p_w1, h_mirror=True)
    b_bl = bezier_bend_t.ref(port_id="0", position=p_w0, rotation=180)

    c.add([b_tr, b_br, b_tl, b_bl])

    c.add_port("W0", port=b_bl.ports["1"])
    c.add_port("W1", port=b_tl.ports["1"])
    c.add_port("E0", port=b_br.ports["1"])
    c.add_port("E1", port=b_tr.ports["1"])

    c.info["min_bend_radius"] = bezier_bend_t.info["min_bend_radius"]

    for pname, p in c.ports.items():
        if p.port_type != "optical":
            c.add_port(pname, port=p)

    return c


if __name__ == "__main__":
    from pp.components import coupler

    # component = coupler(gap=0.3)
    # component = pp.c.mzi2x2(with_elec_connections=True)

    c = package_optical2x2(coupler)
    print(c.ports["E1"].y - c.ports["E0"].y)
    print(c.ports)
    pp.show(c)



import hashlib
from itertools import count
from typing import Dict, List, Tuple, Union
from pp.component import Component


class SequenceGenerator:
    def __init__(
        self, start_sequence="IL", repeated_sequence="ASASBSBS", end_sequence="LO"
    ):
        """
        A simple sequence generator.
        Main use case: any type of cascade of components with repeating patterns
        such as serpentine, cutbacks etc...

        Component sequences have two ports by default. In case you'd need to access other ports, component sequences automatically add aliases for the components forming the sequence.
        They use the component local name and append a suffix index starting from 1, so you may access the ports from any subcomponent. But I don't think this is needed here

        Usually we can break these components in 3 parts:
        - there is a starting pattern with input and possibly some special
        connections
        - then a repeating pattern
        - An ending pattern with an output


        Example of symbol meaning

        A: bend connected with input W0
        B: bend connected with input N0
        I: taper with input '1'
        O: taper with input '2'
        S: short straight waveguide
        L: long straight waveguide
        """

        self.start_sequence = start_sequence
        self.end_sequence = end_sequence
        self.repeated_sequence = repeated_sequence

    def get_sequence(self, n=2):
        return self.start_sequence + n * self.repeated_sequence + self.end_sequence


def _parse_component_name(name: str) -> Tuple[str, bool]:
    """
    If the component name has more than one character and starts with "!"
    then we need to flip along the axis given by the input port angle
    """
    if len(name) != 1 and name[0] == "!":
        return (name[1:], True)
    return (name, False)


def _flip_ref(c_ref, port_name):
    a = c_ref.ports[port_name].angle
    if a in [0, 180]:
        c_ref.reflect_v(port_name)
    else:
        c_ref.reflect_h(port_name)
    return c_ref


def component_sequence(
    sequence: List[str],
    string_to_device_in_out_ports: Dict[
        str, Union[Tuple[Component, str, str], Tuple[None, str, str]]
    ],
    ports_map: Dict[str, Tuple[str, str]] = {},
    input_port_name: str = "in",
    output_port_name: str = "out",
    start_orientation: float = 0.0,
    name_prefix: None = None,
) -> Component:
    """
    This generates a component from a sequence and a dictionnary to interprete each
    symbol in the sequence.

    Args:
        sequence: a string or a list of symbols
        string_to_device_in_out_ports: maps symbols to (device, input, output)
        ports_map: (optional) extra port mapping using the convention
            {port_name: (alias_name, port_name)}

    Returns:
        component containing the sequence of sub-components
    instantiated and connected together in the sequence order

    """
    # Remove all None devices from the sequence
    sequence = sequence[:]
    to_rm = []
    for i, d in enumerate(sequence):
        _name_device, _ = _parse_component_name(d)
        _device, _, _ = string_to_device_in_out_ports[_name_device]
        if _device is None:
            to_rm += [i]

    while to_rm:
        sequence.pop(to_rm.pop())

    # To generate unique aliases for each instance
    counters = {k: count(start=1) for k in string_to_device_in_out_ports.keys()}

    def _next_id(name):
        return "{}{}".format(name, next(counters[name]))

    component = Component()

    # Add first device and input port
    name_start_device, do_flip = _parse_component_name(sequence[0])
    _input_device, input_port, prev_port = string_to_device_in_out_ports[
        name_start_device
    ]

    prev_device = component.add_ref(_input_device, alias=_next_id(name_start_device))

    if do_flip:
        prev_device = _flip_ref(prev_device, input_port)

    prev_device.rotate(angle=start_orientation)

    component.add_port(name=input_port_name, port=prev_device.ports[input_port])

    # Generate and connect all elements from the sequence
    for s in sequence[1:]:
        s, do_flip = _parse_component_name(s)

        _device, input_port, next_port = string_to_device_in_out_ports[s]

        device = component.add_ref(_device, alias=_next_id(s))

        if do_flip:
            device = _flip_ref(device, input_port)

        device.connect(input_port, prev_device.ports[prev_port])
        prev_device = device
        prev_port = next_port

    # Deal with edge case where the sequence contains only one component
    if len(sequence) == 1:
        device = prev_device
        next_port = prev_port

    # Add output port
    try:
        component.add_port(name=output_port_name, port=device.ports[next_port])
    except:
        print(sequence)
        raise

    # Add any extra port specified in ports_map
    for name, (alias, alias_port_name) in ports_map.items():
        component.add_port(name=name, port=component[alias].ports[alias_port_name])

    if name_prefix is not None:
        _md5 = hashlib.md5()
        _md5.update(str(string_to_device_in_out_ports).encode())
        _md5.update(str(sequence).encode())
        component.name = "{}_{}".format(name_prefix, _md5.hexdigest())
    return component


if __name__ == "__main__":
    import pp

    bend180 = pp.c.bend_circular180()
    wg_heater = pp.c.waveguide_heater()
    wg = pp.c.waveguide()

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (bend180, "W0", "W1"),
        "B": (bend180, "W1", "W0"),
        "H": (wg_heater, "W1", "E1"),
        "-": (wg, "W0", "E0"),
    }

    # Generate a sequence
    # This is simply a chain of characters. Each of them represents a component
    # with a given input and and a given output

    sequence = "AB-H-H-H-H-BA"
    component = pp.c.component_sequence(sequence, string_to_device_in_out_ports)
    pp.qp(component)
    pp.show(component)
